# -*- coding: utf-8 -*-

###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################


class Genbank:
    """ Genbank file format parser.
	This parser is based on the format description presented on the NCIBI website
	https://www.ncbi.nlm.nih.gov/Sitemap/samplerecord.html
	"""

    authorized_keywords = [
        "LOCUS",
        "DEFINITION",
        "ACCESSION",
        "VERSION",
        "KEYWORDS",
        "SOURCE",
        "REFERENCE",
        "COMMENT",
        "FEATURES",
        "ORIGIN",
    ]

    def __init__(self, filename):
        self.filename = filename
        self.sequence = {}
        self.parsing_function = None

    def read(self):
        with open(self.filename) as reader:
            for line in reader:
                line = " ".join(line.split())

                # Empty line
                if len(line) == 0:
                    continue

                # Sequence separator found
                if line.startswith("//"):
                    if len(self.sequence):
                        yield self.sequence
                    self.sequence = {}
                    continue

                self.parse_routing(line)

        if len(self.sequence) > 0:
            yield self.sequence

    # --------------- Parser router ---------------

    def parse_routing(self, line):
        # Find the right line parser
        if line.startswith("LOCUS"):
            self.parsing_function = self.parse_locus
        elif line.startswith("DEFINITION"):
            self.parsing_function = self.parse_definition
        elif line.startswith("ACCESSION"):
            self.parsing_function = self.parse_accession
        elif line.startswith("VERSION"):
            self.parsing_function = self.parse_version
        elif line.startswith("KEYWORDS"):
            self.parsing_function = self.parse_keywords
        elif line.startswith("SOURCE"):
            self.parsing_function = self.parse_source
        elif line.startswith("REFERENCE"):
            self.parsing_function = self.parse_reference
        elif line.startswith("COMMENT"):
            self.parsing_function = self.parse_comment
        elif line.startswith("FEATURES"):
            self.parsing_function = self.parse_feature
        elif line.startswith("ORIGIN"):
            self.parsing_function = self.parse_origin
        # TODO: Remove this elif when all the parser is over
        elif line[: line.find(" ")] in Genbank.authorized_keywords:
            self.parsing_function = None
            pass

        # Parse the line
        if self.parsing_function != None:
            self.parsing_function(line)
        else:
            print("Impossible to parse line\n{}".format(line))

    # --------------- Sub-parser by keyword ---------------

    def parse_locus(self, line):
        """ Parse the LOCUS line.
		The line is splited into 5 values as follow :
		- Sequence id
		- Sequence length
		- Molecule type (can be multiple values. Please read the genbank documentation for values)
		- Genbank division
		- Modification date
		"""
        # split the locus line
        split = line.split()
        # Add false values for if the number of values is incorect
        for _ in range(len(split), 7):
            split += ["None"]

        # Parse values of interest
        id, length = split[1:3]
        type = " ".join(split[4:-2])
        division, date = split[-2:]

        # Create dictionary that will contain all the locus values
        self.sequence["LOCUS"] = {
            "id": id,
            "length": length,
            "mol_type": type,
            "genbank_div": division,
            "date": date,
        }

    def parse_definition(self, line):
        """ Parse the DEFINITION section.
		This section contains a text without format restriction.
		"""
        # first call
        if not "DEFINITION" in self.sequence:
            # remove the DEFINITION keyword from the text
            self.sequence["DEFINITION"] = line[line.find(" ") + 1 :]
        else:
            self.sequence["DEFINITION"] = "{} {}".format(
                self.sequence["DEFINITION"], line
            )

    def parse_accession(self, line):
        """ Parse the text as a uniq accession number"""
        split = line.split()

        # First line of the accession
        if split[0] == "ACCESSION":
            # Accession id
            self.sequence["ACCESSION"] = {"id": split[1]}

            # Other values from the accession field
            if len(split) > 2:
                self.sequence["ACCESSION"]["other"] = " ".join(split[2:])
        # In case of multi line
        elif "ACCESSION" in self.sequence and "other" in self.sequence["ACCESSION"]:
            self.sequence["ACCESSION"]["other"] = "{} {}".format(
                self.sequence["ACCESSION"]["other"], " ".join(split)
            )

    def parse_version(self, line):
        """ Parse the text as the version number.
		This value should begin by the accession id a point and then the sequence version number
		Just after the version id the GI identifier can be parsed too
		"""
        split = line.split()

        # First line of the version
        if split[0] == "VERSION":
            # Version id
            self.sequence["VERSION"] = {"id": split[1]}

            # Geninfo Identifier
            if len(split) > 2 and split[2].startswith("GI"):
                self.sequence["VERSION"]["GI"] = split[2][split[2].find(":") + 1 :]
                del split[2]

            # Other values from the Version field
            if len(split) > 2:
                self.sequence["VERSION"]["other"] = " ".join(split[2:])
        # In case of multi line
        elif "VERSION" in self.sequence and "other" in self.sequence["VERSION"]:
            self.sequence["VERSION"]["other"] = "{} {}".format(
                self.sequence["VERSION"]["other"], " ".join(split)
            )

    def parse_keywords(self, line):
        """ This field is deprecated
		See the documentation for more info
		"""
        if not line.startswith("KEYWORDS ."):
            self.sequence["KEYWORDS"] = line[line.find(" ") + 1 :]

    def parse_source(self, line):
        """ Parse the SOURCE label.
		Just after the SOURCE keyword there is a short description.
		The complete description is expected after the optional keyword ORGANISM
		"""
        if line.startswith("SOURCE"):
            self.sequence["SOURCE"] = {"short": line[line.find(" ") + 1 :]}
        elif line.startswith("ORGANISM"):
            self.parsing_function = self.parse_organism
            self.parse_organism(line)
        else:
            self.sequence["SOURCE"]["short"] = "{} {}".format(
                self.sequence["SOURCE"]["short"], line
            )

    def parse_organism(self, line):
        """ Parse the source organism.
		After the ORGANISM keyword the name of the organism is expected.
		The folowing lines are dedicated to the lineage description
		"""
        if line.startswith("ORGANISM"):
            self.sequence["SOURCE"]["ORGANISM"] = {"name": line[line.find(" ") + 1 :]}
        else:
            if "lineage" in self.sequence["SOURCE"]["ORGANISM"]:
                self.sequence["SOURCE"]["ORGANISM"]["lineage"] = "{} {}".format(
                    self.sequence["SOURCE"]["ORGANISM"]["lineage"], line
                )
            else:
                self.sequence["SOURCE"]["ORGANISM"]["lineage"] = line

    reference_keywords = ["AUTHORS", "TITLE", "JOURNAL", "PUBMED", "REMARK"]

    def parse_reference(self, line):
        """ Parse the publication sections
		The keyword REFERENCE can appear as much as you want
		"""
        # Add the publication list
        if not "REFERENCE" in self.sequence:
            self.sequence["REFERENCE"] = []

        # create new entry
        if line.startswith("REFERENCE"):
            self.ref = {}
            self.sequence["REFERENCE"].append(self.ref)
            return

        # dispatch new values
        keyword = line[: line.find(" ")]
        if keyword == "AUTHORS":
            self.parsing_function = self.parse_authors
        elif keyword == "TITLE":
            self.parsing_function = self.parse_title
        elif keyword == "JOURNAL":
            self.parsing_function = self.parse_journal
        elif keyword == "PUBMED":
            self.parsing_function = self.parse_pubmed
        elif keyword == "REMARK":
            self.parsing_function = self.parse_remark
        else:
            print("Genbank: impossible to parse the following line:\n{}".format(line))
            return

        # Parse keywords for reference
        self.parsing_function(line)

    def is_in_the_correct_subtree(self, line, expected_keyword):
        """ Verify if is in the right subtree of the reference parser.
		If not, parse with the right subtree and return False
		"""
        keyword = line[: line.find(" ")]
        # if must go in another branch of the reference parsing
        if keyword == expected_keyword:
            return True

        if keyword in Genbank.reference_keywords:
            self.parsing_function = self.parse_reference
            self.parsing_function(line)
            return False

        return True

    def parse_authors(self, line):
        if not self.is_in_the_correct_subtree(line, "AUTHORS"):
            return

        # Parse the authors
        if line.startswith("AUTHORS"):
            self.ref["AUTHORS"] = line[line.find(" ") + 1 :]
        else:
            self.ref["AUTHORS"] = "{} {}".format(self.ref["AUTHORS"], line)

    def parse_title(self, line):
        if not self.is_in_the_correct_subtree(line, "TITLE"):
            return

        # Parse the title
        if line.startswith("TITLE"):
            self.ref["TITLE"] = line[line.find(" ") + 1 :]
        else:
            self.ref["TITLE"] = "{} {}".format(self.ref["TITLE"], line)

    def parse_journal(self, line):
        if not self.is_in_the_correct_subtree(line, "JOURNAL"):
            return

        # Parse the journal
        if line.startswith("JOURNAL"):
            self.ref["JOURNAL"] = line[line.find(" ") + 1 :]
        else:
            self.ref["JOURNAL"] = "{} {}".format(self.ref["JOURNAL"], line)

    def parse_pubmed(self, line):
        if not self.is_in_the_correct_subtree(line, "PUBMED"):
            return

        # Parse the pubmed id
        if line.startswith("PUBMED"):
            self.ref["PUBMED"] = line[line.find(" ") + 1 :]
        else:
            print("Genbank parser: unexpected multiline pubmed id")

    def parse_remark(self, line):
        if not self.is_in_the_correct_subtree(line, "REMARK"):
            return

        # Parse the remark line
        if line.startswith("REMARK"):
            self.ref["REMARK"] = line[line.find(" ") + 1 :]
        else:
            self.ref["REMARK"] = "{} {}".format(self.ref["REMARK"], line)

    def parse_comment(self, line):
        """ Parse the COMMENT field as a text
		"""
        if line.startswith("COMMENT"):
            self.sequence["COMMENT"] = line[line.find(" ") + 1 :]
        else:
            self.sequence["COMMENT"] = "{} {}".format(self.sequence["COMMENT"], line)

    def parse_origin(self, line):
        """ Parse the origin label and next lines to construct the sequence """
        # Define the sequence on the label ORIGIN
        if line.startswith("ORIGIN"):
            self.sequence["ORIGIN"] = ""

        # parse the sequence
        else:
            # Extract the sub sequence from the line
            sub_seq = "".join(line.split()[1:])
            # Add the subsequence to the complete sequence
            self.sequence["ORIGIN"] = "{}{}".format(self.sequence["ORIGIN"], sub_seq)

    # --------------- Feature parsing ---------------

    # List of feature keywords
    features = [
        "assembly_gap",
        "C_region",
        "CDS",
        "centromere",
        "D-loop",
        "D_segment",
        "exon",
        "gap",
        "gene",
        "iDNA",
        "intron",
        "J_segment",
        "mat_peptide",
        "misc_binding",
        "misc_difference",
        "misc_feature",
        "misc_recomb",
        "misc_RNA",
        "misc_structure",
        "mobile_element",
        "modified_base",
        "mRNA",
        "ncRNA",
        "N_region",
        "old_sequence",
        "operon",
        "oriT",
        "polyA_site",
        "precursor_RNA",
        "prim_transcript",
        "primer_bind",
        "propeptide",
        "protein_bind",
        "regulatory",
        "repeat_region",
        "rep_origin",
        "rRNA",
        "S_region",
        "sig_peptide",
        "source",
        "stem_loop",
        "STS",
        "telomere",
        "tmRNA",
        "transit_peptide",
        "tRNA",
        "unsure",
        "V_region",
        "V_segment",
        "variation",
        "3'UTR",
        "5'UTR",
    ]
    # List of qualifiers
    qualifiers = [
        "allele",
        "altitude",
        "anticodon",
        "artificial_location",
        "bio_material",
        "bound_moiety",
        "cell_line",
        "cell_type",
        "chromosome",
        "citation",
        "clone",
        "clone_lib",
        "codon_start",
        "collected_by ",
        "collection_date ",
        "compare",
        "country",
        "cultivar",
        "culture_collection",
        "db_xref",
        "dev_stage",
        "direction",
        "EC_number",
        "ecotype",
        "environmental_sample",
        "estimated_length",
        "exception",
        "experiment",
        "focus",
        "frequency",
        "function",
        "gap_type",
        "gene",
        "gene_synonym",
        "germline",
        "haplogroup",
        "haplotype",
        "host",
        "identified_by ",
        "inference",
        "isolate",
        "isolation_source",
        "lab_host",
        "lat_lon ",
        "linkage_evidence",
        "locus_tag",
        "macronuclear",
        "map",
        "mating_type",
        "mobile_element_type",
        "mod_base",
        "mol_type",
        "ncRNA_class",
        "note",
        "number",
        "old_locus_tag",
        "operon",
        "organelle ",
        "organism",
        "partial",
        "PCR_conditions",
        "PCR_primers",
        "phenotype",
        "plasmid",
        "pop_variant",
        "product",
        "protein_id",
        "proviral",
        "pseudo",
        "pseudogene",
        "rearranged",
        "recombination_class",
        "regulatory_class",
        "replace",
        "ribosomal_slippage",
        "rpt_family",
        "rpt_type",
        "rpt_unit_range",
        "rpt_unit_seq",
        "satellite",
        "segment",
        "serotype",
        "serovar",
        "sex",
        "specimen_voucher",
        "standard_name",
        "strain",
        "sub_clone",
        "submitter_seqid",
        "sub_species",
        "sub_strain",
        "tag_peptide",
        "tissue_lib",
        "tissue_type",
        "transgenic",
        "translation",
        "transl_except",
        "transl_table",
        "trans_splicing ",
        "type_material",
        "variety",
    ]

    def parse_feature(self, line):
        """ Parse the features.
		The list of feature used is extracted from the appendix II of the oficial ref.
		For more informations: http://www.insdc.org/documents/feature_table.html#7.2
		The list of qualifier are also extracted from the official documetation, appendix III
		More informations: http://www.insdc.org/documents/feature_table.html#7.3
		"""
        keyword = line[: line.find(" ")]

        # Global features keyword
        if keyword == "FEATURES":
            self.sequence["FEATURES"] = []

        # Feature keyword
        elif keyword in Genbank.features:
            self.current_feature = {
                "type": keyword,
                "position": line[line.find(" ") + 1 :],
            }
            self.sequence["FEATURES"].append(self.current_feature)

        # Qualifier
        elif (
            line[1 : line.find("=") if "=" in line else len(line)] in Genbank.qualifiers
        ):
            # init values
            qualifier = line[1:]
            value = None

            # parse qualifier and value if the operator is dual
            is_dual_operator = "=" in keyword

            if is_dual_operator:
                qualifier, value = qualifier.split("=")
                # remove quotes from strings
                if value[0] == value[-1] == '"':
                    value = value[1:-1]

            # First time that the qualifier is encountered
            if not qualifier in self.current_feature:
                self.current_feature[qualifier] = value
            else:
                # If there is more than 1 element for this qualifier
                if isinstance(self.current_feature[qualifier], list):
                    self.current_feature[qualifier].append(value)
                # Transform a value into a vector of values
                else:
                    self.current_feature[qualifier] = [
                        self.current_feature[qualifier],
                        value,
                    ]

        # End of the FEATURES part
        elif keyword in Genbank.authorized_keywords:
            self.parsing_function = self.parse_routing
            self.parsing_function(line)

        # Unknown line
        else:
            print("Genbank: impossible to parse the following line:")
            print(line)


# gbk = Genbank("data/test_genbank.gbk")
# nb_seq = 0
# for seq in gbk.read():
# 	nb_seq += 1
# 	print(seq)

# print(nb_seq)



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`FASTQ` to :term:`FASTA` and :term:`QUAL` formats"""
from bioconvert import ConvBase, bioconvert_script
from bioconvert.core.base import ConvArg
from bioconvert.core.decorators import compressor, in_gz
from bioconvert.core.decorators import requires, requires_nothing

from mappy import fastx_read
import mmap


__all__ = ["FASTQ2FASTA_QUAL"]


class FASTQ2FASTA_QUAL(ConvBase):
    """Convert :term:`FASTQ` to :term:`FASTA` and :term:`QUAL` formats

    Method based on Python.

    """

    #: Default value
    _default_method = "python"

    def __init__(self, infile, outfile, *args, **kargs):
        """.. rubric:: constructor

        :param str infile: input fastQ file
        :param str outfile: output fastA filename
        :param str outfile2: output QUAL file

        """
        super(FASTQ2FASTA_QUAL, self).__init__(infile, outfile)

        # Here we need to reset the infile and outfile since we have
        # a 1-to-2 conversion.
        self.infile = infile
        self.outfile = outfile[0]
        self.outfile2 = outfile[1]

    @staticmethod
    def _readfq(fp):  # this is a generator function
        last = None  # this is a buffer keeping the last unprocessed line
        while True:  # mimic closure; is it a bad idea?
            if not last:  # the first record or a record following a fastq
                for l in fp:  # search for the start of the next record
                    if l[0] == "@":  # fastq header line
                        last = l[:-1]  # save this line
                        break
            if not last:
                break
            header, seqs, last = last[1:], [], None
            for l in fp:  # read the sequence
                if l[0] in "@+":
                    last = l[:-1]
                    break
                seqs.append(l[:-1])
            seq, leng, seqs = "".join(seqs), 0, []
            for l in fp:  # read the quality
                seqs.append(l[:-1])
                leng += len(l) - 1
                if leng >= len(seq):  # have read enough quality
                    last = None
                    yield header, seq, "".join(seqs)  # yield a fastq record
                    break

    @requires_nothing
    @compressor
    def _method_python(self, *args, **kwargs):
        """This method is inspired by Readfq coded by Heng Li.

        `original Readfq method <https://github.com/lh3/readfq>`_"""
        with open(self.outfile, "w") as fasta, open(
            self.outfile2, "w"
        ) as quality, open(self.infile, "r") as fastq:
            for (name, seq, qual) in FASTQ2FASTA_QUAL._readfq(fastq):
                fasta.write(">{}\n{}\n".format(name, seq))
                quality.write(">{}\n{}\n".format(name, qual))

    @staticmethod
    def get_IO_arguments():
        yield ConvArg(
            names="input_file",
            default=None,
            type=ConvArg.file,
            help="The path to the file to convert.",
        )
        yield ConvArg(
            names="output_file",
            nargs=2,
            default=None,
            type=ConvArg.file,
            output_argument=True,
            help="The path where the result will be stored.",
        )



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`BAM` format to :term:`JSON` format"""
from bioconvert import ConvBase

import colorlog

from bioconvert.core.decorators import requires

logger = colorlog.getLogger(__name__)


class BAM2JSON(ConvBase):
    """Convert :term:`BAM` format to :term:`JSON` file

    Methods available are based on bamtools [BAMTOOLS]_.

    """

    #: Default value
    _default_method = "bamtools"

    def __init__(self, infile, outfile):
        """.. rubric:: constructor

        :param str infile:
        :param str outfile:
        """
        super(BAM2JSON, self).__init__(infile, outfile)

    @requires("bamtools")
    def _method_bamtools(self, *args, **kwargs):
        """Do the conversion :term:`BAM` -> :term:`JSON` using bamtools.

        `BAMTools documentation <https://hcc.unl.edu/docs/applications/app_specific/bioinformatics_tools/data_manipulation_tools/bamtools/running_bamtools_commands/>`_"""

        cmd = "bamtools convert -format json -in {0} -out {1}".format(
            self.infile, self.outfile
        )
        self.execute(cmd)



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################

"""Convert :term:`MAF` file to :term:`SAM` format"""
from bioconvert import ConvBase

import colorlog

logger = colorlog.getLogger(__name__)

__all__ = ["MAF2SAM"]


class MAF2SAM(ConvBase):
    """This is the Multiple alignment format or MIRA assembly format

    This is not Mutation Annotation Format (somatic)


    pbsim creates this kind of data

    Some references:

    - https://github.com/peterjc/maf2sam/
    - https://github.com/arq5x/nanopore-scripts/master/maf-convert.py
    - http://bioperl.org/formats/alignment_formats/MAF_multiple_alignment_format.html

    Those two codes were in Py2 at the time of this implementation. We re-used
    some of the information from maf-convert but the code in
    bioconvert.io.maf can be considered original. 
    """

    #: Default value
    _default_method = "python"

    def __init__(self, infile, outfile):
        super().__init__(infile, outfile)

    def _method_python(self, *args, **kwargs):
        """Internal module
        
        `MAF documentation <https://bioconvert.readthedocs.io/en/refactoring/ref_io.html#module-bioconvert.io.maf>`_"""
        from bioconvert.io import maf

        conv = maf.MAF(self.infile, self.outfile)
        conv.to_sam()



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""misc utility functions """
import os
import sys
import bioconvert
from bioconvert.core.extensions import extensions


__all__ = ["get_extension", "get_format_from_extension", "generate_outfile_name"]


def get_extension(filename, remove_compression=False):
    """Return extension of a filename

    ::

        >>> get_extension("test.fastq")
        fastq
        >>> get_extension("test.fastq.gz")
        fastq

    """
    compression = [".gz", ".bz2", ".bzip2", ".dsrc"]
    if remove_compression is True:
        # remove the .gz, .bz2, ..., extensions
        for this in compression:
            if filename.endswith(this):
                filename = filename.rsplit(".", 1)[0]

    if len(filename.split(".")) == 1:
        return None
    else:
        return os.path.splitext(filename)[-1].lstrip(".")


def generate_outfile_name(infile, out_extension):
    """simple utility to replace the file extension with the given one.

    :param str infile: the path to the Input file
    :param str out_extension: Desired extension
    :return: The file path with the given extension
    :rtype: str
    """
    return "{}.{}".format(os.path.splitext(infile)[0], out_extension)


def get_format_from_extension(extension):
    """get format from extension.

    :param extension: the extension
    :return: the corresponding format
    :rtype: str
    """
    extension = extension.lstrip(".")
    for fmt, ext_list in extensions.items():
        if extension in ext_list:
            return fmt.upper()

    # The extension was not found
    bioconvert.logger.critical(
        "No format was found for extension '{}'".format(extension)
    )
    bioconvert.logger.critical(
        "Use --formats to know the available formats and --help" " for examples"
    )
    sys.exit(1)


def compressor(infile, comp_ext, threads=4):
    # FIXME: could be a method in ConvBase

    if comp_ext == ".gz":
        _log.info("Compressing output into .gz")
        shell("pigz -f -p {} {}".format(threads, infile))
    elif comp_ext == ".bz2":
        _log.info("Compressing output into .bz2")
        shell("pbzip2 -f -p{} {}".format(threads, infile))
    elif comp_ext == ".dsrc":  # !!! only for FastQ files
        _log.info("Compressing output into .dsrc")
        shell("dsrc c -t{} {} {}.dsrc".format(inst.threads, inst.outfile, inst.infile))



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################

""""Convert :term:`CRAM` to :term:`BAM` format"""
import re
import os

from bioconvert import ConvBase
from easydev.multicore import cpu_count

import colorlog

from bioconvert.core.decorators import requires_nothing, requires

logger = colorlog.getLogger(__name__)


class SAM2PAF(ConvBase):
    """Convert :term:`SAM` file to :term:`PAF` file

    The :term:`SAM` and :term:`PAF` formats are described in the :ref:`formats`
    section.

    Description:

    The header of the SAM file (lines starting with @) are dropped. However, the
    length of the target is retrieved from the @SQ line that must be present.


    Consider this SAM file with two alignements only. One is aligned on the
    target (first) while the other is not (indicated by the ``*`` characters)::

        @SQ	SN:ENA|K01711|K01711.1	LN:15894
        @PG	ID:minimap2	PN:minimap2	VN:2.5-r572	CL:minimap2 -a measles.fa Hm2_GTGAAA_L005_R1_001.fastq.gz
        HISEQ:426:C5T65ACXX:5:2302:1943:2127	0	ENA|K01711|K01711.1	448	60	101M	*	00	CTTACCTTCGCATCAAGAGGTACCAACATGGAGGATGAGGCGGACCAATACTTTTCACATGATGATCCAATTAGTAGTGATCAATCCAGGTTCGGATGGTT	BCCFFFFFHHHHHIIJJJJJJIIJJJJJJJJFHIHIJJJIJIIIIGHFFFFFFEEEEEEEDDDDDFDDDDDDDDD>CDDEDEEDDDDDDCCDDDDDDDDCD	NM:i:0	ms:i:202	AS:i:202	nn:i:0	tp:A:P	cm:i:14	s1:i:94	s2:i:0
        HISEQ:426:C5T65ACXX:5:2302:4953:2090	4	*	0	0	*	*	0	0	AGATCGGAAGAGCACACGTCTGAACTCCAGTCACGTGAAAATCTCGTATGCCGTCTTCTGCTTGAAAAAAAAAAAAAAAACAACCAAAAAGAGACGAACAA	CCCFFDDFAFFBHJHGGGIHIJBGGHIIJJJJJJJHGEIJGIFIIIHCBGHIJIIIIIJJHHHHEF@D@=;=,0)0&5&))+(((+((((&+(((()&&)(


    The equivalent PAF file is ::

        HISEQ:426:C5T65ACXX:5:2302:1943:2127	101	0	101	+	ENA|K01711|K01711.1	15894	447	548	101	101	60	NM:i:0	ms:i:202	AS:i:202	nn:i:0	tp:A:P	cm:i:14	s1:i:94	s2:i:0	cg:Z:101M

    In brief, the sequences are dropped. The final file is therefore smaller.
    Extra fields (starting from NM:i:0) can be dropped or kept using the
    keep_extra_field argument. Alignement with ``*`` characters are dropped.
    The first line (@SQ) is used to retrieve the length of the contigs that is
    stored in the PAF file (column 6).


    The 12 compulsary PAF fields are:

    ====== ======== ===========================================
    Col     Type    Description
    ====== ======== ===========================================
    1      string   Query sequence name
    2       int     Query sequence length
    3       int     Query start (0-based)
    4       int     Query end (0-based)
    5       char    Relative strand: "+" or "-"
    6      string   Target sequence name
    7       int     Target sequence length
    8       int     Target start on original strand (0-based)
    9       int     Target end on original strand (0-based)
    10      int     Number of residue matches
    11      int     Alignment block length
    12      int     Mapping quality (0-255; 255 for missing)
    ====== ======== ===========================================


    For developesr:

    Get the measles data from Sequana library (2 paired fastq files)::

        minimap2 measles.fa R1.fastq > approx-mapping.paf

    You can ask minimap2 to generate CIGAR at the cg tag of PAF with:

        minimap2 -c measles.fa R1.fastq > alignment.paf

    or to output alignments in the SAM format::

        minimap2 -a measles.fa R1.fastq > alignment.sam


    The SAM lines must contains 11 positional element and the NM:i and nn:i
    fields (see example above).

    """

    #: Default value
    _default_method = "python"

    def __init__(self, infile, outfile, *args, **kargs):
        """.. rubric:: constructor

        :param str infile: input SAM file
        :param str outfile: output PAF filename

        :reference: This function is a direct translation of 
            https://github.com/lh3/miniasm/blob/master/misc/sam2paf.js (Dec. 
            2017).

        """
        super(SAM2PAF, self).__init__(infile, outfile, *args, **kargs)

    @requires_nothing
    def _method_python(self, *args, **kwargs):
        """Internal method"""
        pattern = r"(\d+)([MIDSHNX=])"

        extra_fields = kwargs.get("extra_fields", "SAM")
        # TODO: what is this ?
        pri_only = kwargs.get("pri_only", True)

        skipped = 0
        reference_lengths = {}
        with open(self.infile, "r") as fin:
            with open(self.outfile, "w") as fout:
                for lineno, line in enumerate(fin.readlines()):

                    if line.startswith("@"):
                        if line.startswith("@SQ"):
                            match = re.findall(r"\tSN:(\S+)", line)
                            name = match[0] if len(match) else "unknown_reference"

                            match = re.findall(r"\tLN:(\d+)", line)
                            if len(match) == 1:
                                reference_lengths[name] = int(match[0])
                            else:
                                raise ValueError(
                                    "Could not parse SQ line to extract the length "
                                    "(LN: field missing maybe ?)"
                                )
                        continue

                    t = line.split()
                    flag = int(t[1])

                    if t[9] != "*" and t[10] != "*" and len(t[9]) != len(t[10]):
                        raise ValueError(
                            "ERROR at line "
                            + str(lineno)
                            + ":inconsistent SEQ and QUAL lengths - "
                            + str(len(t[9]))
                            + " != "
                            + str(len(t[10]))
                        )

                    if t[2] == "*" or (flag & 4):
                        skipped += 1
                        continue

                    # if flag is 256 and pri_only, we skip the alignment
                    if pri_only and (flag & 0x100):
                        continue

                    # Get the reference length for this alignment
                    if t[2] in reference_lengths:
                        tlen = reference_lengths[t[2]]
                    else:
                        raise KeyError(
                            "can't find the length of contig {}".format(t[2])
                        )

                    # The reference is known but the length is not
                    if tlen == -1:
                        raise ValueError(
                            "ERROR at line "
                            + str(lineno)
                            + ": can't find the length of contig "
                            + str(t[2])
                        )

                    # TODO explain what are the nn and NM tags
                    match = re.findall(r"\tnn:i:(\d+)", line)
                    if match:
                        nn = int(match[0])
                    else:
                        nn = 0

                    match = re.findall(r"\tNM:i:(\d+)", line)
                    if match:
                        NM = int(match[0])
                        have_NM = True
                    else:
                        NM = 0
                        have_NM = False
                    NM += nn

                    # See sequana.cigar for more information
                    clip = [0, 0]
                    I = [0, 0]  # Insertion
                    D = [0, 0]  # Deletion
                    M, N = 0, 0  # Matches
                    ql, tl, mm = (
                        0,
                        0,
                        0,
                    )
                    ext_cigar = False
                    n_cigar = 0

                    Zacc = ""
                    for count, letter in re.findall(pattern, t[5]):
                        l = int(count)
                        if letter == "M":
                            M += l
                            ql += l
                            tl += l
                            ext_cigar = False
                            Zacc += "{}M".format(count)
                        elif letter == "I":
                            I[0] += 1
                            I[1] += l
                            ql += l
                            Zacc += "{}I".format(count)
                        elif letter == "D":
                            D[0] += 1
                            D[1] += l
                            tl += l
                            Zacc += "{}D".format(count)
                        elif letter == "N":
                            N += l
                            tl += l
                        elif letter == "S":
                            clip[0 if M == 0 else 1] = l
                            ql += l
                        elif letter == "H":
                            clip[0 if M == 0 else 1] = l
                        elif letter == "=":
                            M += l
                            ql += l
                            tl += l
                            ext_cigar = True
                        elif letter == "X":
                            M += l
                            ql += l
                            tl += l
                            mm += l
                            ext_cigar = True
                        n_cigar += 1

                    prefix_msg = "at line {}: ".format(lineno) + "{}"

                    if n_cigar > 65535:
                        logger.warning(
                            prefix_msg.format(str(n_cigar) + " CIGAR operations")
                        )

                    if tl + int(t[3]) - 1 > tlen:
                        logger.warning(
                            prefix_msg.format(
                                "alignment end "
                                "position larger than ref length; skipped"
                            )
                        )
                        continue

                    if t[9] != "*" and len(t[9]) != ql:
                        logger.warning(
                            prefix_msg.format(
                                " SEQ length inconsistent with CIGAR("
                                + str(len(t[9]))
                                + " != "
                                + str(ql)
                                + "); skipped"
                            )
                        )
                        continue

                    if have_NM is False or ext_cigar:
                        NM = I[1] + D[1] + mm

                    if NM < I[1] + D[1] + mm:
                        logger.warning(
                            prefix_msg.format(
                                " NM is less than the total number of gaps ("
                                + str(NM)
                                + " < "
                                + str(I[1] + D[1] + mm)
                                + ")"
                            )
                        )
                        NM = I[1] + D[1] + mm

                    # extra information to store in the PAF after the 12
                    # extra field from original code. The insert and
                    # deletions (io/in and do/di) and mm is the number of other
                    # substitutions ? NM -I[1] -D[1]
                    extra = [
                        "mm:i:" + str(NM - I[1] - D[1]),
                        "io:i:" + str(I[0]),
                        "in:i:" + str(I[1]),
                        "do:i:" + str(D[0]),
                        "dn:i:" + str(D[1]),
                    ]

                    match = M - (NM - I[1] - D[1])
                    blen = M + I[1] + D[1]
                    qlen = M + I[1] + clip[0] + clip[1]

                    # What does flag 16 means ?
                    if flag & 16:
                        qs = clip[1]
                        qe = qlen - clip[0]
                    else:
                        qs = clip[0]
                        qe = qlen - clip[1]

                    ts = int(t[3]) - 1
                    te = ts + M + D[1] + N

                    ## WARNING: difference with sam2paf.js : we add and substract nn
                    ## from match and blen to agree with the output SAM file
                    ## generated by minimap2

                    # The 12 compulsary fields to have a valid PAF format
                    a = [
                        t[0],
                        qlen,
                        qs,
                        qe,
                        "-" if flag & 16 else "+",
                        t[2],
                        tlen,
                        ts,
                        te,
                        match + nn,
                        blen - nn,
                        t[4],
                    ]
                    # cast to string and save in file
                    a = [str(x) for x in a]

                    # What extra fields do we want to add ?
                    # original fields found in the SAM file ?
                    if extra_fields == "SAM" and len(t) > 11:
                        fout.write(
                            "\t".join(a)
                            + "\t"
                            + "\t".join(t[11:])
                            + "\tcg:Z:{}\n".format(Zacc)
                        )
                    elif extra_fields == "summary":
                        fout.write("\t".join(a) + "\t" + "\t".join(extra) + "\n")
                    elif extra_fields is None:
                        fout.write("\t".join(a) + "\n")

        self.skipped = skipped

    # @requires(external_binaries=["k8", "paftools"])
    # def _method_paftools(self, *args, **kwargs):
    #    cmd = "paftools sam2paf {} > {}".format(self.infile, self.outfile)
    #    self.execute(cmd)



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`YAML` to :term:`JSON` format"""
import yaml
import json
from bioconvert import ConvBase
import colorlog

from bioconvert.core.decorators import requires_nothing
from bioconvert.core.decorators import compressor

logger = colorlog.getLogger(__name__)

__all__ = ["YAML2JSON"]


class YAML2JSON(ConvBase):
    """Convert :term:`YAML` file into :term:`JSON` file

    Conversion is based on yaml and json standard Python modules

    .. note:: YAML comments will be lost in JSON output

    :reference: http://yaml.org/spec/1.2/spec.html#id2759572
    """

    #: Default value
    _default_method = "python"

    def __init__(self, infile, outfile, *args, **kargs):
        """.. rubric:: constructor

        :param str infile: input YAML file.
        :param str outfile: input JSON file
        """
        super(YAML2JSON, self).__init__(infile, outfile, *args, **kargs)

    @requires_nothing
    @compressor
    def get_json(self):
        """Return the JSON dictionary corresponding to the YAML input."""
        try:
            data = yaml.load(open(self.infile, "r"), Loader=yaml.FullLoader)
        except: #pragma: no cover
            data = yaml.load(open(self.infile, "r"))

        return json.dumps(data, sort_keys=True, indent=4)

    @requires_nothing
    @compressor
    def _method_python(self, *args, **kwargs):
        """Internal method"""
        with open(self.outfile, "w") as outfile:
            outfile.write(self.get_json())



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`BPLINK` to :term:`VCF` format"""
import colorlog
import os

from bioconvert import ConvBase
from bioconvert.core.decorators import requires
from bioconvert.core.utils import generate_outfile_name

_log = colorlog.getLogger(__name__)


class BPLINK2VCF(ConvBase):
    """Converts a genotype dataset bed+bim+fam in :term:`BPLINK` format to
    vcf :term:`VCF` format

    Conversion is based on plink [PLINK]_ executable.

    .. warning:: **plink** takes several inputs and outputs and does not need
        extensions. What is required is a prefix. Bioconvert usage is therefore::

            bioconvert bplink2vcf plink_toy plink_toy.vcf

        Since there is no extension, you must be explicit by providing the
        conversion name (bplink2plink). This command will search for 3 input 
        files plink_toy.bed, plink_toy.bim and plink_toy.fam. It will then 
        create two output files named plink_toy.ped and plink_toy.map

    """

    #: Default value
    _default_method = "plink"

    def __init__(self, infile, outfile=None, *args, **kwargs):
        """.. rubric:: constructor

        :param str infile: input :term:`BPLINK` files.
        :param str outfile: (optional) output :term:`VCF` file
        """
        if not outfile:
            outfile = generate_outfile_name(infile, "vcf")
        super(BPLINK2VCF, self).__init__(infile, outfile)

    @requires("plink")
    def _method_plink(self, *args, **kwargs):
        """Convert using plink executable.

        `plink documentation <http://hpc.ilri.cgiar.org/beca/training/data_mgt_2017/BackgroundMaterial/PlinkTutorial.pdf>`_"""
        outfile = self.outfile
        if os.path.splitext(outfile)[1] == ".vcf":
            outfile = os.path.splitext(outfile)[0]
        cmd = "plink --bfile {infile} --recode vcf --out {outfile}".format(
            infile=self.infile, outfile=outfile
        )
        self.execute(cmd)



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`BAM` format to :term:`FASTA` format"""
from bioconvert import ConvBase
from bioconvert.core.decorators import requires
from bioconvert.core.utils import get_extension
import subprocess
import os
import itertools


class BAM2FASTA(ConvBase):
    """Convert sorted :term:`BAM` file into :term:`FASTA` file

    Methods available are based on samtools [SAMTOOLS]_ or bedtools [BEDTOOLS]_.

    .. warning:: Using the bedtools method, the R1 and R2 reads must be next to
        each other so that the reads are sorted similarly

    .. warning:: there is no guarantee that the R1/R2 output file are sorted
        similarly in paired-end case due to supp and second reads

    """

    #: Default value
    _default_method = "samtools"

    def __init__(self, infile, outfile):
        """.. rubric:: constructor

        :param str infile: BAM file
        :param str outfile: FASTA file

        """
        super(BAM2FASTA, self).__init__(infile, outfile)

    """@requires("bamtools")
    def __method_bamtools(self, *args, **kwargs):
        #  fastq are split on several lines (80 characters)
        # this method contains supplementary reads and we don't know
        # what to do with them for now. So, this method is
        # commented. Indeed final R1 and R2 files will not be paired.

        cmd = "bamtools stats -in '%s' | sed '12!d' | awk '{print $3}' " % (self.infile)
        ps = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,universal_newlines=True)
        isPaired = ps.communicate()[0].strip()

        # Collect the extension
        ext = os.path.splitext(self.outfile)[1]

        cmd = "bamtools convert -format fasta -in {} -out {}".format(
            self.infile, self.outfile)
        self.execute(cmd)
   """

    @requires("samtools")
    def _method_samtools(self, *args, **kwargs):
        """
        do the conversion :term:`BAM` -> :term:`FASTA` using samtools.

        .. note:: fasta are on one line

        `SAMtools documentation <http://www.htslib.org/doc/samtools.html>`_"""
        # Test if input bam file is paired
        p = subprocess.Popen(
            "samtools view -c -f 1 {}".format(self.infile).split(),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )

        isPaired = p.communicate()[0].strip()

        # Collect the extension
        ext = os.path.splitext(self.outfile)[1]

        # FIXME: this compression code may be factorised ?

        output_ext = get_extension(self.outfile, remove_compression=True)

        # If the output file extension is compress extension
        if ext in [".gz", ".bz2"]:
            outbasename = os.path.splitext(self.outfile)[0].split(".", 1)[0]

            if ext == ".gz":
                compresscmd = "gzip -f"
            elif ext == ".bz2":
                compresscmd = "pbzip2 -f"

            # When the input file is not paired and the output file needs to be compressed
            if isPaired == "0":
                cmd = "samtools fasta {} > {}.{}".format(
                    self.infile, outbasename, output_ext
                )
                self.execute(cmd)
                cmd = "{} {}.{}".format(compresscmd, outbasename, output_ext)
                self.execute(cmd)
            # When the input file is paired and the output file needs to be compressed
            else:
                cmd = "samtools fasta -1 {}_1.{} -2 {}_2.{} -n {} ".format(
                    outbasename, output_ext, outbasename, output_ext, self.infile
                )
                self.execute(cmd)
                cmd = "{} {}_1.{}".format(compresscmd, outbasename, output_ext)
                self.execute(cmd)
                cmd = "{} {}_2.{}".format(compresscmd, outbasename, output_ext)
                self.execute(cmd)
        else:
            outbasename = os.path.splitext(self.outfile)[0]

            # When the input file is not paired
            if isPaired == "0":
                cmd = "samtools fasta {} > {}".format(self.infile, self.outfile)
                self.execute(cmd)
            # When the input file is paired
            else:
                cmd = "samtools fasta -1 {}_1.{} -2 {}_2.{} -n {} ".format(
                    outbasename, output_ext, outbasename, output_ext, self.infile
                )
                self.execute(cmd)



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
""".. rubric:: Standalone application dedicated to conversion"""
import os

import colorlog

from bioconvert.core.base import ConvMeta
from bioconvert.core.registry import Registry

_log = colorlog.getLogger(__name__)

from bioconvert.core.base import make_chain
from bioconvert.core.utils import get_extension as getext
from bioconvert.core.utils import get_format_from_extension
import sys

__all__ = ["Bioconvert"]


class Bioconvert(object):
    """Universal converter used by the standalone

    ::

        from bioconvert import Bioconvert
        c = Bioconvert("test.fastq", "test.fasta", threads=4, force=True)


    """

    def __init__(self, infile, outfile, force=False, threads=None, extra=None):
        """.. rubric:: constructor

        :param str infile: The path of the input file.
        :param str outfile: The path of The output file
        :param bool force: overwrite output file if it exists already
            otherwise raises an error

        """

        # check existence of output file. If it exists,
        # fails except if force argument is set to True

        if type(outfile) is str:
            outfile = [outfile]

        if type(infile) is str:
            infile = [infile]

        # some checking on the output files (existence, special case of dsrc)
        for filename in outfile:
            if os.path.exists(filename) is True:
                msg = "output file {} exists already.".format(filename)
                if force is False:
                    _log.critical(
                        "output file exists. If you are using bioconvert, use --force "
                    )
                    raise ValueError(msg)
                else:
                    _log.warning(msg + " --force used so will be over written")

            # Only fastq files can be compressed with dsrc
            if filename.endswith(".dsrc"):
                # only valid for FastQ files extension
                # dsrc accepts only .fastq file extension
                if filename.endswith(".fastq.dsrc") is False:
                    msg = (
                        "When compressing with .dsrc extension, "
                        + "only files ending with .fastq extension are "
                        + "accepted. This is due to the way dsrc executable "
                        + "is implemented."
                    )
                    _log.critical(msg)
                    raise IOError

        Lin = len(infile)
        Lout = len(outfile)

        self.inext = []
        self.outext = []

        # populate the inext
        for filename in infile:
            # example: fastq.gz to fasta.bz2
            # Here, we want to decompress, convert, compress.
            # so we need the extension without .gz or .bz2
            # We should have inext set to fastq and outext
            # set to fasta.bz2
            self.inext.append(getext(filename, remove_compression=True))

        # populate the outext
        for filename in outfile:
            self.outext.append(getext(filename, remove_compression=True))

        # special case one to one for compression/decompression
        # Case 2, fastq.gz to fastq.bz2
        # data is not changed, just the type of compression, so we want
        # to keep the original extensions, here inext and outext  will contain
        # .gz and .bz2
        # if 1 to 1 and same extension, we overwrite self.inext and self.outext
        if Lin == Lout == 1:
            if self.inext == self.outext:
                _log.info("decompression/compression mode")
                self.inext = [getext(infile[0])]
                self.outext = [getext(outfile[0])]

        self.mapper = Registry()

        # From the input parameters 1 and 2, we get the module name
        if not list(
            set(list(self.mapper.get_converters_names())).intersection(sys.argv)
        ):
            # get format from extensions
            in_fmt = [get_format_from_extension(x) for x in self.inext]
            out_fmt = [get_format_from_extension(x) for x in self.outext]
        else:
            in_fmt, out_fmt = ConvMeta.split_converter_to_format(
                list(
                    set(list(self.mapper.get_converters_names())).intersection(sys.argv)
                )[0]
            )

        self.in_fmt = in_fmt
        self.out_fmt = out_fmt

        self.in_fmt = [format.lower() for format in in_fmt]
        self.in_fmt = tuple(in_fmt)

        self.out_fmt = [format.lower() for format in out_fmt]
        self.out_fmt = tuple(out_fmt)

        _log.info("Input: {}".format(self.in_fmt))
        _log.info("Output: {}".format(self.out_fmt))

        try:
            class_converter = self.mapper[(self.in_fmt, self.out_fmt)]
            self.name = class_converter.__name__

        except KeyError:
            # This module name was not found
            # Try to find path of converters
            conv_path = self.mapper.conversion_path(self.in_fmt, self.out_fmt)
            _log.debug("path: {}".format(conv_path))
            if conv_path:
                _log.info("Direct conversion not implemented. " "Chaining converters.")
                # implemented in bioconvert/core/base.py
                # using temporary files
                class_converter = make_chain(
                    [(pair, self.mapper[pair]) for pair in conv_path]
                )
            else:
                msg = "Requested input format ('{}') to output format ('{}') is not available in bioconvert".format(
                    self.in_fmt, self.out_fmt,
                )
                _log.critical(msg)
                _log.critical(
                    "Use --formats to know the available formats and --help for examples"
                )
                raise Exception(msg)

        # If --threads provided, we update the threads attribute

        # FIXME: hack for the compression/decompression decorators

        if Lin == 1:
            infile = infile[0]

        if Lout == 1:
            outfile = outfile[0]

        self.converter = class_converter(infile, outfile)
        if threads is not None:
            self.converter.threads = threads
        if extra:
            self.converter._extra_arguments = extra

        _log.info(
            "Using {} class (with {} threads if needed)".format(
                self.converter.name, self.converter.threads
            )
        )

        # For the benchmarking only
        self.converter.others = {}

    def __call__(self, *args, **kwargs):
        self.converter(*args, **kwargs)

    # def boxplot_benchmark(self, *args, **kwargs):
    #    results = self.converter.boxplot_benchmark(*args, **kwargs)
    #    return results



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""List of formats and associated extensions"""
from easydev import AttrDict

# Formats can be of type
# - sequence
# - alignment
# - binary
# - compression
# - database
# - variant

#: List of formats and their extensions included in Bioconvert
extensions = {
    "abi": ["abi", "ab1"],  # sequence
    "agp": ["agp"],  # assembly
    "bam": ["bam"],  # alignment
    "bcf": ["bcf"],  # variant
    "bed": ["bed"],  # database
    "bedgraph": ["bedgraph", "bg"],  # database
    "bigwig": ["bigwig", "bw"],  # database
    "bigbed": ["bb", "bigbed"],
    "bz2": ["bz2"],  # compression
    "bplink": ["bplink"],
    "cdao": ["cdao"],  # phylo
    "cram": ["cram"],  # alignment
    "clustal": ["clustal", "aln", "clw"],  # phylo
    "cov": ["cov"],  # coverage (chrom name,  pos, depth)
    "csv": ["csv"],  # database
    "dsrc": ["dsrc"],  # compression
    "embl": ["embl"],  # annotation/sequence
    "ena": ["ena"],
    "faa": ["faa", "mpfa"],  # fasta multiple amino acid
    "fasta": ["fasta", "fa", "fst"],  # sequence
    "fastq": ["fastq", "fq"],  # sequence
    "genbank": ["genbank", "gbk", "gb"],  # annotation/sequence
    "gfa": ["gfa"],  # assembly
    "gff2": ["gff"],
    "gff3": ["gff3"],  # annotation
    "gz": ["gz"],
    "json": ["json"],  # database
    "maf": ["maf"],  # !! this is MIRA format, not mutation alignment format
    "newick": ["newick", "nw", "nhx", "nwk"],  # phylo
    "nexus": ["nexus", "nx", "nex", "nxs"],  # phylo
    "ods": ["ods"],  # database
    "paf": ["paf"],  # assembly
    "pdb": ["pdb"],
    "phylip": ["phy", "ph", "phylip"],  # phylo
    "phyloxml": ["phyloxml", "xml"],  # phylo
    "plink": ["plink"],
    "qual": ["qual"],  # seauence
    "sam": ["sam"],  # alignement
    "scf": ["scf"],  # alignement
    "sra": ["sra"],  # sra format
    "stockholm": ["sto", "sth", "stk", "stockholm"],  # alignment
    "twobit": ["2bit"],  # sequence
    "tsv": ["tsv"],  # database
    "vcf": ["vcf"],  # variant
    "wiggle": ["wig", "wiggle"],
    "wig": ["wig"],
    "xls": ["xls"],  # database
    "xlsx": ["xlsx"],  # database
    "xmfa": ["xmfa"],
    "yaml": ["yaml", "YAML"],  # database
}

extensions = AttrDict(**extensions)



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Simplified version of shell.py module from snakemake package

Temporary module that would probably need some checking / simplification

This is a temporary replacement to the execute() in convbase
that also uses subprocess.Popen but is slower
"""
import _io
import sys
import os
import subprocess as sp

import colorlog

_log = colorlog.getLogger(__name__)


STDOUT = sys.stdout


class shell:
    _process_args = {}
    _process_prefix = ""
    _process_suffix = ""

    @classmethod
    def executable(cls, cmd):
        if os.path.split(cmd)[-1] == "bash":
            cls._process_prefix = "set -euo pipefail; "
        cls._process_args["executable"] = cmd

    @classmethod
    def prefix(cls, prefix):
        cls._process_prefix = prefix

    @classmethod
    def suffix(cls, suffix):
        cls._process_suffix = suffix

    def __new__(cls, cmd, *args, is_async=False, iterable=False, read=False, **kwargs):
        if "stepout" in kwargs:
            raise KeyError("Argument stepout is not allowed in shell command.")

        stdout = sp.PIPE if iterable or is_async or read else STDOUT

        close_fds = sys.platform != "win32"

        _log.debug(cmd)
        proc = sp.Popen(
            "{} {} {}".format(cls._process_prefix, cmd.rstrip(), cls._process_suffix),
            bufsize=-1,
            shell=True,
            stdout=stdout,
            close_fds=close_fds,
            **cls._process_args
        )

        ret = None
        if iterable:
            return cls.iter_stdout(proc, cmd)
        if read:
            ret = proc.stdout.read()
        elif is_async:
            return proc
        retcode = proc.wait()
        if retcode:
            raise sp.CalledProcessError(retcode, cmd)
        return ret

    @staticmethod
    def iter_stdout(proc, cmd):
        for l in proc.stdout:
            yield l[:-1].decode()
        retcode = proc.wait()
        if retcode:
            raise sp.CalledProcessError(retcode, cmd)



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`FASTA` to :term:`PHYLIP` format"""
import colorlog
from bioconvert import ConvBase
from bioconvert.core.decorators import requires
from bioconvert.core.decorators import compressor
from Bio import SeqIO

_log = colorlog.getLogger(__name__)


__all__ = ["FASTA2PHYLIP"]


class FASTA2PHYLIP(ConvBase):
    """Converts a sequence alignment in :term:`FASTA` format to :term:`PHYLIP` format

    Conversion is based on Bio Python modules

    Methods available are based on squizz [SQUIZZ]_ or biopython [BIOPYTHON]_ or 
    goalign [GOALIGN]_. Squizz is the default 
    (https://github.com/bioconvert/bioconvert/issues/149). Phylip created is a
    strict phylip that is with 10 characters on the first column.

    """

    #: Default value
    _default_method = "biopython"

    def __init__(self, infile, outfile=None, alphabet=None, *args, **kwargs):
        """.. rubric:: constructor

        :param str infile: input :term:`FASTA` file.
        :param str outfile: (optional) output :term:`PHYLIP` file
        """
        super(FASTA2PHYLIP, self).__init__(infile, outfile)

    @requires(python_library="biopython")
    @compressor
    def _method_biopython(self, *args, **kwargs):
        """For this method we use the biopython package Bio.SeqIO. 
        
        `Bio.SeqIO Documentation <https://biopython.org/docs/1.76/api/Bio.SeqIO.html>`_"""
        sequences = list(SeqIO.parse(self.infile, "fasta"))
        count = SeqIO.write(sequences, self.outfile, "phylip")
        _log.debug("Converted %d records to phylip" % count)

    @requires("squizz")
    @compressor
    def _method_squizz(self, *args, **kwargs):
        """Convert fasta file in Phylip interleaved format using squizz tool.
        The fasta file must be an alignement file, this means that all sequences must
        have the same length (with the gap) otherwise an error will be raised."""
        cmd = "squizz -c PHYLIPI {infile} > {outfile}".format(
            infile=self.infile, outfile=self.outfile
        )
        self.execute(cmd)

    @requires("go")
    @compressor
    def _method_goalign(self, *args, **kwargs):
        """Convert fasta file in Phylip interleaved format using goalign tool.

        `goalign documentation <https://github.com/fredericlemoine/goalign>`_
        
        The fasta file must be an alignemnt file, this means that  all sequences 
        must have the same length (with the gap) otherwise an error will be raised
        """
        self.install_tool("goalign")
        cmd = "goalign reformat phylip -i {infile} -o {outfile}".format(
            infile=self.infile, outfile=self.outfile
        )
        self.execute(cmd)



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`BAM` format to :term:`COV` format"""
from bioconvert import ConvBase
import colorlog

from bioconvert.core.decorators import requires

_log = colorlog.getLogger(__name__)

__all__ = ["BAM2COV"]


class BAM2COV(ConvBase):
    """Convert sorted :term:`BAM` file into :term:`COV` file


    Note that the COV format is of the form::

        chr1    1   0
        chr1    2   0
        chr1    3   0
        chr1    4   0
        chr1    5   0

    that is contig name, position, coverage.

    .. warning:: the BAM file must be sorted. This can be achieved with
        bamtools using *bamtools sort -in INPUT.bam*

    Methods available are based on samtools [SAMTOOLS]_ or bedtools [BEDTOOLS]_.
    """

    #: Default value
    # _default_method = "samtools"
    _default_method = "bedtools"

    def __init__(self, infile, outfile):
        """.. rubric:: Constructor

        :param str infile: The path to the input BAM file. **It must be sorted**.
        :param str outfile: The path to the output file
        """
        super().__init__(infile, outfile)

    @requires("samtools")
    def _method_samtools(self, *args, **kwargs):
        """Do the conversion sorted :term:`BAM` -> :term:`BED` using samtools

        `SAMtools documentation <http://www.htslib.org/doc/samtools.html>`_"""
        cmd = "samtools depth -aa {} > {}".format(self.infile, self.outfile)
        self.execute(cmd)

    @requires("bedtools")
    def _method_bedtools(self, *args, **kwargs):
        """Do the conversion sorted :term:`BAM` -> :term:`BED` using bedtools

        `bedtools documentation <https://bedtools.readthedocs.io/en/latest/>`_"""
        cmd = "bedtools genomecov -d -ibam {} > {}".format(self.infile, self.outfile)
        self.execute(cmd)



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018-2019  Institut Pasteur, Paris and CNRS.                #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Utility used by the script bioconvert_init to initiate a new plugin"""
import os

__all__ = ["InitConverter"]


template = '''
# -*- coding: utf-8 -*-
###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018-2019  Institut Pasteur, Paris and CNRS.                #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
""" description """

from bioconvert.core.base import ConvArg
from bioconvert.core.decorators import requires, requires_nothing
from bioconvert import ConvBase

__all__ = ["{input}2{output}"]


class {input}2{output}(ConvBase):
    """Convert :term:`{input}` file to :term:`{output}` file

    Some description to be added by the developer

    """

    _default_method = "default"

    def __init__(self, infile, outfile, *args, **kargs):
        """.. rubric:: constructor

        :param str infile: input {input} file
        :param str outfile: output {output} filename

        """
        super({input}2{output}, self).__init__(infile, outfile, *args, **kargs)

    @requires("ls")
    def _method_default(self, *args, **kwargs):
        """some description"""
        cmd = "ls"
        # use self.infile, self.outfile
        self.execute(cmd)
'''


class InitConverter:
    """Class to create a new plugin based on a simple template

    If the input/output formats are not known by bioconvert (not available in
    the module core/extensions.py then, the developer will need to add two
    attributes manually::

        input_ext = ["yourextension"]
        output_ext = ["yourextension"]

    We recommand to use the script bioconvert_init

    """

    def __init__(self, inputext, outputext):
        self.input = inputext
        self.output = outputext

    def get_content(self):
        return template.format(
            input=self.input.upper(),
            output=self.output.upper(),
            inputext=self.input,
            outputext=self.output,
        )



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""simple levenshtein implementation in pure Python """
# from https://pypi.python.org/pypi/pylev#downloads


def classic_levenshtein(string_1, string_2):
    """
    Calculates the Levenshtein distance between two strings.

    This version is easier to read, but significantly slower than the version
    below (up to several orders of magnitude). Useful for learning, less so
    otherwise.

    Usage::

        >>> classic_levenshtein('kitten', 'sitting')
        3
        >>> classic_levenshtein('kitten', 'kitten')
        0
        >>> classic_levenshtein('', '')
        0

    """
    len_1 = len(string_1)
    len_2 = len(string_2)
    cost = 0

    if len_1 and len_2 and string_1[0] != string_2[0]:
        cost = 1

    if len_1 == 0:
        return len_2
    elif len_2 == 0:
        return len_1
    else:
        return min(
            classic_levenshtein(string_1[1:], string_2) + 1,
            classic_levenshtein(string_1, string_2[1:]) + 1,
            classic_levenshtein(string_1[1:], string_2[1:]) + cost,
        )


def wf_levenshtein(string_1, string_2):
    """
    Calculates the Levenshtein distance between two strings.

    This version uses the Wagner-Fischer algorithm.

    Usage::

        >>> wf_levenshtein('kitten', 'sitting')
        3
        >>> wf_levenshtein('kitten', 'kitten')
        0
        >>> wf_levenshtein('', '')
        0

    """
    len_1 = len(string_1) + 1
    len_2 = len(string_2) + 1

    d = [0] * (len_1 * len_2)

    for i in range(len_1):
        d[i] = i
    for j in range(len_2):
        d[j * len_1] = j

    for j in range(1, len_2):
        for i in range(1, len_1):
            if string_1[i - 1] == string_2[j - 1]:
                d[i + j * len_1] = d[i - 1 + (j - 1) * len_1]
            else:
                d[i + j * len_1] = min(
                    d[i - 1 + j * len_1] + 1,  # deletion
                    d[i + (j - 1) * len_1] + 1,  # insertion
                    d[i - 1 + (j - 1) * len_1] + 1,  # substitution
                )

    return d[-1]



__version__ = "0.1.2"
import pkg_resources

try:
    version = pkg_resources.require("bioconvert")[0].version
except:
    version = __version__

import os
import colorlog

# This will create a HOME/.config/bioconvert where files (e.g., executables)
# can be downloaded
from easydev import CustomConfig

configuration = CustomConfig("bioconvert", verbose=True)

os.environ["GOPATH"] = os.environ["HOME"] + "/go"
os.environ["PATH"] = os.environ["GOPATH"] + "/bin/:" + os.environ["PATH"]

from easydev.logging_tools import Logging

logger = Logging("bioconvert", "INFO")


def bioconvert_script(filename, where=None):
    bioconvert_path = bioconvert.__path__[0]
    share = os.path.join(bioconvert_path, "misc")
    if where:
        filename = os.path.join(share, where, filename)
    else:
        filename = os.path.join(share, filename)
    if not os.path.exists(filename):
        raise FileNotFoundError("unknown file {}".format(filename))
    return filename


def bioconvert_data(filename, where=None):
    """Simple utilities to retrieve data sets from bioconvert/data directory

    :param str filename: the name of the data file to get the path
    :param str where:
    """
    bioconvert_path = bioconvert.__path__[0]
    share = os.path.join(bioconvert_path, "data")
    # in the code one may use / or \
    if where:
        filename = os.path.join(share, where, filename)
    else:
        filename = os.path.join(share, filename)
    if not os.path.exists(filename):
        raise FileNotFoundError("unknown file {}".format(filename))
    return filename


def info():
    from bioconvert.core.registry import Registry

    r = Registry()
    info = r.get_info()
    converters = [x for x in info.items()]
    data = [info[k] for k, v in info.items()]
    msg = "Bioconvert contains {} converters including {} methods"
    return msg.format(len(converters), sum(data))


import bioconvert
from bioconvert.core.base import ConvBase
from bioconvert.core.registry import Registry
from bioconvert.core.decorators import requires, compressor
from bioconvert.core.benchmark import Benchmark
from bioconvert.core.converter import Bioconvert
from bioconvert.core.shell import shell



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Converts :term:`PHYLOXML` file to :term:`NEXUS` format."""
import colorlog

from bioconvert import ConvBase
from bioconvert.core.decorators import requires
from bioconvert.core.decorators import compressor

_log = colorlog.getLogger(__name__)


__all__ = ["PHYLOXML2NEXUS"]


class PHYLOXML2NEXUS(ConvBase):
    """
    Converts a tree file from :term:`PHYLOXML` format to :term:`NEXUS` format.

    Methods available are based on gotree [GOTREE]_.

    """

    #: Default value
    _default_method = "gotree"

    def __init__(self, infile, outfile=None, *args, **kwargs):
        """.. rubric:: constructor

        :param str infile: input :term:`PHYLOXML` file.
        :param str outfile: (optional) output :term:`NEXUS` file
        """
        super().__init__(infile, outfile)

    @requires("go")
    @compressor
    def _method_gotree(self, *args, **kwargs):
        """Convert :term:`PHYLOXML`  file in :term:`NEXUS` format using gotree tool.

        `gotree documentation <https://github.com/fredericlemoine/gotree>`_"""
        self.install_tool("gotree")
        cmd = "gotree reformat nexus -i {infile} -o {outfile} -f phyloxml"
        cmd = cmd.format(infile=self.infile, outfile=self.outfile)
        self.execute(cmd)



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`CSV` format to :term:`TSV` format"""
import csv
import colorlog

from bioconvert.core.base import ConvArg
from bioconvert.core.decorators import requires, requires_nothing
from bioconvert.core.decorators import compressor, in_gz
from bioconvert.core.base import ConvBase


logger = colorlog.getLogger(__name__)


class CSV2TSV(ConvBase):
    """Convert :term:`CSV` file into :term:`TSV` file

    Available methods: Python, Pandas

    Methods available are based on python or Pandas [PANDAS]_.

    .. seealso:: :class:`~bioconvert.csv2tsv.TSV2CSV`
    """

    #: Default value
    _default_method = "python"
    DEFAULT_IN_SEP = ","
    DEFAULT_OUT_SEP = "\t"
    DEFAULT_LINE_TERMINATOR = "\n"

    def __init__(self, infile, outfile):
        """.. rubric:: Constructor

        :param str infile: comma-separated file
        :param str outfile: tabulated file
        """
        super(CSV2TSV, self).__init__(infile, outfile)

    @requires_nothing
    @compressor
    def _method_python(
        self,
        in_sep=DEFAULT_IN_SEP,
        out_sep=DEFAULT_OUT_SEP,
        line_terminator=DEFAULT_LINE_TERMINATOR,
        *args,
        **kwargs
    ):
        """Do the conversion :term:`CSV` -> :term:`TSV` using standard Python modules.

        `csv documentation <https://docs.python.org/3/library/csv.html>`_"""
        with open(self.infile, "r") as in_stream, open(self.outfile, "w") as out_stream:
            writer = csv.writer(
                out_stream, delimiter=out_sep, lineterminator=line_terminator
            )
            reader = csv.reader(in_stream, delimiter=in_sep)
            for row in reader:
                writer.writerow(row)

    @requires_nothing
    @compressor
    def _method_python_v2(
        self,
        in_sep=DEFAULT_IN_SEP,
        out_sep=DEFAULT_OUT_SEP,
        line_terminator=DEFAULT_LINE_TERMINATOR,
        *args,
        **kwargs
    ):
        """Do the conversion :term:`CSV` -> :term:`CSV` using csv module.

        .. note:: This method cannot escape nor quote output char
        
        `csv documentation <https://docs.python.org/3/library/csv.html>`_"""
        with open(self.infile, "r") as in_stream, open(self.outfile, "w") as out_stream:
            reader = csv.reader(in_stream, delimiter=in_sep)
            for row in reader:
                out_stream.write(out_sep.join(row))
                out_stream.write(line_terminator)

    @requires(python_library="pandas")
    @compressor
    def _method_pandas(
        self,
        in_sep=DEFAULT_IN_SEP,
        out_sep=DEFAULT_OUT_SEP,
        line_terminator=DEFAULT_LINE_TERMINATOR,
        *args,
        **kwargs
    ):
        """Do the conversion :term:`CSV` -> :term:`TSV` using Pandas library
        
        
        `pandas documentation <https://pandas.pydata.org/docs/>`_"""
        import pandas as pd

        pd.read_csv(self.infile, sep=in_sep,).to_csv(
            self.outfile,
            sep=out_sep,
            line_terminator=line_terminator,
            index=False,
            header="infer",
        )

    @classmethod
    def get_additional_arguments(cls):
        yield ConvArg(
            names=["--in-sep",],
            default=cls.DEFAULT_IN_SEP,
            help="The separator used in the input file",
        )
        yield ConvArg(
            names=["--out-sep",],
            default=cls.DEFAULT_OUT_SEP,
            help="The separator used in the output file",
        )
        yield ConvArg(
            names=["--line-terminator",],
            default=cls.DEFAULT_LINE_TERMINATOR,
            help="The line terminator used in the output file",
        )



# -*- coding: utf-8 -*-

###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################

"""
Available methods per converter
=====================================

Plot number of implemented methods per converter.


"""
#################################################
#
from bioconvert.core.registry import Registry

r = Registry()
info = r.get_info()

# The available unique converters
converters = [x for x in info.items()]

# the number of methods per converter
data = [info[k] for k, v in info.items()]


# the number of formats
A1 = [x[0] for x in list(r.get_conversions())]
A2 = [x[1] for x in list(r.get_conversions())]
formats = set(A1 + A2)

print("Number of formats: {}".format(len(formats)))
print("Number of converters: {}".format(len(converters)))
print("Number of methods : {}".format(sum(data)))

#####################################################
from pylab import hist, clf, xlabel, grid

clf()
hist(data, range(17), ec="k", zorder=2, align="left")
xlabel("Number of methods")
grid(zorder=-1)



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Converts :term:`PHYLIP` file to :term:`XMFA` format."""
import colorlog
from Bio import AlignIO

from bioconvert import ConvBase
from bioconvert.core.decorators import requires
from bioconvert.core.decorators import compressor

_log = colorlog.getLogger(__name__)


__all__ = ["PHYLIP2XMFA"]


class PHYLIP2XMFA(ConvBase):
    """
    Converts a sequence alignment from :term:`PHYLIP` format to :term:`XMFA` 

    Methods available are based on biopython [BIOPYTHON]_.

    """

    #: Default value
    _default_method = "biopython"

    def __init__(self, infile, outfile=None, *args, **kwargs):
        """.. rubric:: constructor

        :param str infile: input :term:`PHYLIP` file.
        :param str outfile: (optional) output :term:`NEXUS` file
        """
        super().__init__(infile, outfile)

    @requires(python_libraries=["biopython"])
    @compressor
    def _method_biopython(self, *args, **kwargs):
        """Convert :term:`PHYLIP` interleaved file in :term:`XMFA` (Mauve)format.

        `Bio.AlignIO <https://biopython.org/docs/1.76/api/Bio.AlignIO.html>`_"""
        sequences = list(AlignIO.parse(self.infile, "phylip"))
        count = AlignIO.write(sequences, self.outfile, "mauve")
        _log.info("Converted %d records to phylip" % count)



##############################################################################
#  This file is part of Bioconvert software
#
#  Copyright (c) 2017 - Bioconvert Development Team
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/biokit/bioconvert
#  documentation: http://bioconvert.readthedocs.io
##############################################################################
"""convert :term:`CSV` to :term:`XLS` format"""
import csv

import colorlog

from bioconvert import ConvBase
from bioconvert.core.decorators import requires, requires_nothing
from bioconvert.core.decorators import compressor, in_gz
from bioconvert.core.base import ConvArg

logger = colorlog.getLogger(__name__)


__all__ = ["CSV2XLS"]


class CSV2XLS(ConvBase):
    """Convert :term:`CSV` file to :term:`XLS` file

    Methods available are based on python, pyexcel [PYEXCEL]_,
    or pandas [PANDAS]_.

    """

    #: Default value
    _default_method = "pandas"
    DEFAULT_IN_SEP = ","
    DEFAULT_LINE_TERMINATOR = "\n"
    DEFAULT_SHEET_NAME = "Sheet 1"

    def __init__(self, infile, outfile, *args, **kargs):
        """.. rubric:: constructor

        :param str infile: input CSV file
        :param str outfile: output XLS filename

        """
        super(CSV2XLS, self).__init__(infile, outfile, *args, **kargs)

    @requires(python_libraries=["pyexcel", "pyexcel-xls"])
    @compressor
    def _method_pyexcel(
        self, in_sep=DEFAULT_IN_SEP, sheet_name=DEFAULT_SHEET_NAME, *args, **kwargs
    ):
        """Do the conversion :term:`CSV` -> :term:`XLS` using pyexcel modules.

        `pyexcel documentation <http://docs.pyexcel.org/en/latest/>`_"""
        rows = []
        with open(self.infile, "r") as in_stream:
            reader = csv.reader(in_stream, delimiter=in_sep)
            for row in reader:
                rows.append(row)

        from pyexcel_xls import save_data
        from collections import OrderedDict

        data = OrderedDict()
        data.update({sheet_name: rows})
        save_data(self.outfile, data)

    @requires(python_libraries=["pandas"])
    @compressor
    def _method_pandas(
        self, in_sep=DEFAULT_IN_SEP, sheet_name=DEFAULT_SHEET_NAME, *args, **kwargs
    ):
        """Do the conversion :term:`CSV` -> :term:`XLS` using Panda modules.

        `pandas documentation <https://pandas.pydata.org/docs/>`_"""
        import pandas as pd

        writer = pd.ExcelWriter(self.outfile, engine='openpyxl')
        pd.read_csv(self.infile, sep=in_sep, header="infer",).to_excel(
            excel_writer=writer, sheet_name=sheet_name, index=False,
        )
        writer.save()

    @classmethod
    def get_additional_arguments(cls):
        yield ConvArg(
            names=["--sheet-name",],
            default=cls.DEFAULT_SHEET_NAME,
            help="The name of the sheet to create",
        )
        yield ConvArg(
            names=["--in-sep",],
            default=cls.DEFAULT_IN_SEP,
            help="The separator used in the input file",
        )



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNT General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`FASTA` format to :term:`FAA` format"""
import textwrap
import colorlog

from bioconvert import ConvBase
from bioconvert.core.decorators import compressor
from bioconvert.core.decorators import requires
from bioconvert.io.fasta import Fasta

_log = colorlog.getLogger(__name__)


__all__ = ["FASTA2FAA"]


class FASTA2FAA(ConvBase):
    """

    Methods available is a bioconvert implementation.

    """

    #: Default value
    _default_method = "bioconvert"

    def __init__(self, infile, outfile):
        """
        :param str infile: The path to the input FASTA file
        :param str outfile: The path to the output FASTQ file
        """
        super(FASTA2FAA, self).__init__(infile, outfile)

        self.codons = {
            "ATA": "I",
            "ATC": "I",
            "ATT": "I",
            "ATG": "M",
            "ACA": "T",
            "ACC": "T",
            "ACG": "T",
            "ACT": "T",
            "AAC": "N",
            "AAT": "N",
            "AAA": "K",
            "AAG": "K",
            "AGC": "S",
            "AGT": "S",
            "AGA": "R",
            "AGG": "R",
            "CTA": "L",
            "CTC": "L",
            "CTG": "L",
            "CTT": "L",
            "CCA": "P",
            "CCC": "P",
            "CCG": "P",
            "CCT": "P",
            "CAC": "H",
            "CAT": "H",
            "CAA": "Q",
            "CAG": "Q",
            "CGA": "R",
            "CGC": "R",
            "CGG": "R",
            "CGT": "R",
            "GTA": "V",
            "GTC": "V",
            "GTG": "V",
            "GTT": "V",
            "GCA": "A",
            "GCC": "A",
            "GCG": "A",
            "GCT": "A",
            "GAC": "D",
            "GAT": "D",
            "GAA": "E",
            "GAG": "E",
            "GGA": "G",
            "GGC": "G",
            "GGG": "G",
            "GGT": "G",
            "TCA": "S",
            "TCC": "S",
            "TCG": "S",
            "TCT": "S",
            "TTC": "F",
            "TTT": "F",
            "TTA": "L",
            "TTG": "L",
            "TAC": "Y",
            "TAT": "Y",
            "TAA": "_",
            "TAG": "_",
            "TGC": "C",
            "TGT": "C",
            "TGA": "_",
            "TGG": "W",
        }

    @compressor
    def _method_bioconvert(self, *args, **kwargs):
        """Internal method."""
        data = Fasta(self.infile)
        with open(self.outfile, "w") as fout:
            for line in data.read():
                fout.write(">{}\t{}\n".format(line["id"], line["comment"]))
                seq = line["value"]

                # translate the sequence
                aa = "".join(
                    (
                        self.codons.get(seq[i : i + 3], "X")
                        for i in range(0, len(seq), 3)
                    )
                )
                fout.write("\n".join(textwrap.wrap(aa, 60)) + "\n")



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`BAM` format to :term:`BEDGRAPH` format"""
from bioconvert import ConvBase
import colorlog

from bioconvert.core.decorators import requires

_log = colorlog.getLogger(__name__)


__all__ = ["BAM2BEDGRAPH"]


class BAM2BEDGRAPH(ConvBase):
    """Convert sorted :term:`BAM` file into :term:`BEDGRAPH` file

    Compute the coverage (depth) in BEDGRAPH.
    Regions with zero coverage are also reported.


    Note that this BEDGRAPH format is of the form::

        chrom chromStart chromEnd dataValue

    Note that consecutive positions with same values are compressed.

    ::

        chr1    0   75  0
        chr1    75  176 1
        chr1    176  177 2


    .. warning:: the BAM file must be sorted. This can be achieved with
        bamtools.


    Methods available are based on bedtools [BEDTOOLS]_ and mosdepth
    [MOSDEPTH]_.
    """

    # 4 minutes with bedtools and 20s with mosdepth
    #: Default value
    _default_method = "bedtools"
    _threading = True

    def __init__(self, infile, outfile):
        """.. rubric:: Constructor

        :param str infile: The path to the input BAM file.
            **It must be sorted**.
        :param str outfile: The path to the output file
        """
        super().__init__(infile, outfile)

    @requires("bedtools")
    def _method_bedtools(self, *args, **kwargs):
        """Do the conversion using bedtools.

        `bedtools documentation <https://bedtools.readthedocs.io/en/latest/>`_"""
        cmd = "bedtools genomecov -bga -ibam {} > {}".format(self.infile, self.outfile)
        self.execute(cmd)

    @requires("mosdepth")
    def _method_mosdepth(self, *args, **kwargs):
        """Do the conversion using mosdepth.

        `mosdepth documentation <https://github.com/brentp/mosdepth>`_"""
        # For testing, we need to save into a specific temporary directory
        import tempfile

        with tempfile.TemporaryDirectory() as tmpdir:
            try:
                cmd = "mosdepth {}/.bioconvert -t {}  {}".format(
                    tmpdir, self.threads, self.infile
                )
                self.execute(cmd)

                if self.outfile.endswith(".gz"):
                    pass
                else:
                    cmd = "gunzip -c {}/.bioconvert.per-base.bed.gz > {}".format(
                        tmpdir, self.outfile
                    )
                    self.execute(cmd)
            except Exception as err:
                raise (err)
            finally:
                cmd = "rm -f {name}/.bioconvert.per-base.bed.gz {name}/.bioconvert.per-base.bed.gz.csi"
                cmd += " {name}/.bioconvert.mosdepth.global.dist.txt"
                self.execute(cmd.format(name=tmpdir))



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`BEDGRAPH` to :term:`BIGWIG` format"""
import os
import colorlog

from bioconvert.core.base import ConvArg, ConvBase

from bioconvert.core.decorators import requires

_log = colorlog.getLogger(__name__)


__all__ = ["BEDGRAPH2BIGWIG"]


class BEDGRAPH2BIGWIG(ConvBase):
    """Converts :term:`BEDGRAPH` format to :term:`BIGWIG` format

    Conversion is based on bedGraph2BigWig tool. Note that an 
    argument --chrom-sizes is required.

    """

    #: Default value
    _default_method = "ucsc"

    def __init__(self, infile, outfile):  # , alphabet=None, *args, **kwargs):
        """.. rubric:: constructor

        :param str infile: input :term:`BEDGRAPH` file.
        :param str outfile: (optional) output :term:`BIGWIG` file
        """
        super(BEDGRAPH2BIGWIG, self).__init__(infile, outfile)

    @requires("bedGraphToBigWig")
    def _method_ucsc(self, *args, **kwargs):
        """Convert bedgraph file in bigwig format using ucsc tool.

        `bigWig documentation <https://genome.ucsc.edu/goldenpath/help/bigWig.html>`_

        `chromosome size <http://hgdownload.cse.ucsc.edu/goldenPath/hg38/bigZips/hg38.chrom.sizes>`_"""
        chrom_sizes = kwargs.get("chrom_sizes", None)
        if chrom_sizes is None:
            raise ValueError("Must provide --chrom-sizes option")

        cmd = "bedGraphToBigWig {infile}  {chrom_sizes} {outfile}".format(
            infile=self.infile, outfile=self.outfile, chrom_sizes=chrom_sizes
        )
        self.execute(cmd)

    @classmethod
    def get_additional_arguments(cls):
        yield ConvArg(
            names="--chrom-sizes",
            default=None,
            help="a two-column file/URL: <chromosome name> <size in bases>",
        )



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`FASTA` format to :term:`FASTQ` format"""
from bioconvert import ConvBase
from bioconvert.core.base import ConvArg
import colorlog
import sys
from bioconvert.core.extensions import extensions

from bioconvert.core.decorators import requires

_log = colorlog.getLogger(__name__)


class FASTA_QUAL2FASTQ(ConvBase):
    """Convert FASTA and QUAL back into a FASTQ file

    Method based on pysam [PYSAM]_.

    """

    #: Default value
    _default_method = "pysam"

    def __init__(self, infile, outfile):
        """
        :param list infile: The path to the input FASTA file, the path to the input QUAL file
        :param str outfile: The path to the output FASTQ file
        """
        super(FASTA_QUAL2FASTQ, self).__init__(infile, outfile)

    @requires(python_library="pysam")
    def _method_pysam(self, *args, **kwargs):
        """This method uses the FastxFile function of the Pysam python module.
        
        `FastxFile documentation <https://pysam.readthedocs.io/en/latest/api.html#pysam.FastxFile.close>`_"""
        from pysam import FastxFile

        if self.infile[1] is None:
            _log.error("No quality file provided. Please add a quality file path ")
            sys.exit(1)

        else:  # length must be equal and identifiers sorted similarly
            with open(self.outfile, "w") as fastq_out:
                for seq, qual in zip(
                    FastxFile(self.infile[0]), FastxFile(self.infile[1])
                ):
                    assert seq.name == qual.name
                    if seq.comment:
                        fastq_out.write(
                            "@{0} {1}\n{2}\n+\n{3}\n".format(
                                seq.name, seq.comment, seq.sequence, qual.sequence
                            )
                        )
                    else:
                        fastq_out.write(
                            "@{0}\n{1}\n+\n{2}\n".format(
                                seq.name, seq.sequence, qual.sequence
                            )
                        )

    @staticmethod
    def get_IO_arguments():
        yield ConvArg(
            names="input_file",
            nargs=2,
            default=None,
            type=ConvArg.file,
            help="The path to the file to convert.",
        )
        yield ConvArg(
            names="output_file",
            default=None,
            type=ConvArg.file,
            output_argument=True,
            help="The path where the result will be stored.",
        )



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`CRAM` file to :term:`SAM` format"""
import os
from bioconvert import ConvBase
from bioconvert.core.base import ConvArg

import colorlog

from bioconvert.core.decorators import requires

logger = colorlog.getLogger(__name__)


class CRAM2SAM(ConvBase):
    """Convert :term:`CRAM` file to :term:`SAM` file

    The conversion requires the reference corresponding to the input file
    It can be provided as an argument with the standalone (*-\\-reference*). 
    Otherwise, users are asked to provide it.

    Methods available are based on samtools [SAMTOOLS]_.
    """

    #: Default value
    _default_method = "samtools"
    _threading = True

    def __init__(self, infile, outfile, *args, **kargs):
        """.. rubric:: constructor

        :param str infile: input CRAM file
        :param str outfile: output SAM filename

        """
        super(CRAM2SAM, self).__init__(infile, outfile, *args, **kargs)

    def _get_reference(self):
        # In case the --reference is not used
        msg = "Please enter the reference corresponding "
        msg += "to the input BAM file:"
        reference = input(msg)
        if os.path.exists(reference) is False:
            raise IOError("Reference required")
        else:
            logger.debug("Reference exists ({}).".format(reference))
        return reference

    @requires("samtools")
    def _method_samtools(self, *args, **kwargs):
        """Here we use the SAMtools tool.

        `SAMtools documentation <http://www.htslib.org/doc/samtools.html>`_"""
        # -h means include header in SAM output
        reference = kwargs.get("reference", None)
        if reference is None:
            reference = self._get_reference()

        cmd = "samtools view -@ {} -h -T {} {} > {}".format(
            self.threads, reference, self.infile, self.outfile
        )
        self.execute(cmd)



# -*- coding: utf-8 -*-

###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
import re


class Gff3:
    """Read a gff v3 file

    See the format description at
    https://github.com/The-Sequence-Ontology/Specifications/blob/master/gff3.md


    """

    def __init__(self, filename):
        self.filename = filename

    def read(self):
        """ Read annotations one by one creating a generator """
        with open(self.filename) as reader:
            line = None

            for line in reader:

                # Skip metadata and comments
                if line.startswith("#"):
                    continue

                # Skip empty lines
                if not line.strip():
                    continue

                # Format checking
                split = line.rstrip().split("\t")

                if len(split) < 9:
                    # Wrong line format
                    if len(split) > 0:
                        print(
                            "Impossible to read the following line regarding the gff3 specifications"
                        )
                        print(line)
                    continue

                annotation = self.process_main_fields(split[0:8])
                annotation["attributes"] = self.process_attributes(split[8])

                yield annotation

    def process_main_fields(self, fields):
        annotation = {}

        # Unique id of the sequence
        annotation["seqid"] = self.decode_small(fields[0])
        # Optional source
        if fields[1] != ".":
            annotation["source"] = self.decode_small(fields[1])
        # Annotation type
        annotation["type"] = self.decode_small(fields[2])
        # Start and stop
        annotation["start"] = int(fields[3])
        annotation["stop"] = int(fields[4])
        # Optional score field
        if fields[5] != ".":
            annotation["score"] = float(fields[5])
        # Strand
        if fields[6] == "+" or fields[6] == "-" or fields[6] == "?" or fields[6] == ".":
            annotation["strand"] = fields[6]
        # Phase
        if fields[7] != ".":
            annotation["phase"] = int(fields[7]) % 3

        return annotation

    def process_attributes(self, text):
        attributes = {}

        # split into mutliple attributes
        split = text.split(";")
        for attr in split:
            # find the separator
            idx = attr.find("=")

            # parse tags and associated values
            value = self.decode_complete(attr[idx + 1 :])
            if len(value) == 1:
                value = value[0]
            attributes[self.decode_complete(attr[:idx])] = value

        return attributes

    @staticmethod
    def decode_small(text):
        # Tabulation
        text = re.sub("%09", "\t", text)
        # newline
        text = re.sub("%0A", "\n", text)
        # return
        text = re.sub("%0D", "\r", text)
        # percent
        text = re.sub("%25", "%", text)

        return text

    @staticmethod
    def decode_complete(text):
        text = Gff3.decode_small(text)
        # semicolon
        text = re.sub("%3B", ";", text)
        # equals
        text = re.sub("%3D", "=", text)
        # ampersand
        text = re.sub("%26", "&", text)
        # comma
        text = re.sub("%2C", ",", text)

        return text


# if __name__ == "__main__":
#     file = "/home/yoann/Projects/Hub/bioconvert/bioconvert/data/gff3_example.gff"
#     gff = Gff3(file)
#     for annotation in gff.read():
#         print(annotation)



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018-2019  Institut Pasteur, Paris and CNRS.                #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
from itertools import groupby
import math


class MAFLine(object):
    """A reader for :term:`MAF` format.

    mode refname start algsize strand refsize alignment

    ::

        a
        s ref    100 10 + 100000 ---AGC-CAT-CATT
        s contig 0   10 + 10     ---AGC-CAT-CATT

        a
        s ref    100 12 + 100000 ---AGC-CAT-CATTTT
        s contig 0   12 + 12     ---AGC-CAT-CATTTT

    The alignments are stored by pair, one item for the reference, one for the
    query. The query (second line) starts at zero.

    """

    def __init__(self, line):
        self._items = line.split()
        assert self._items[0] in "aspq", self._items

    @property
    def name(self):
        return self._items[1]

    @property
    def sequence_size(self):
        return int(self._items[5])

    @property
    def mode(self):
        return self._items[0]

    @property
    def strand(self):
        return self._items[4]

    @property
    def alignment_start(self):  # alignment start
        return int(self._items[2])

    @property
    def alignment_size(self):
        return int(self._items[3])

    @property
    def alignment(self):
        return self._items[6]

    def get_alignment_length(self):
        return len(self._items[6].replace("-", ""))


class MAF(object):
    """A reader for :term:`MAF` format."""

    def __init__(self, filename, outfile=None):
        self.filename = filename
        self.outfile = outfile

    def count_insertions(self, alnString):
        """return length without insertion, forward and reverse shift"""
        gaps = alnString.count("-")
        forwardFrameshifts = alnString.count("\\")
        reverseFrameshifts = alnString.count("/")
        letters = len(alnString) - gaps - forwardFrameshifts - reverseFrameshifts
        return letters, forwardFrameshifts, reverseFrameshifts

    #    @SQ SN:NC_002929    LN:4086189
    #    @PG ID:bioconvert VN:?? CL:bioconvert input.maf output.sam

    def to_sam(self):
        # identifier flag ref start qual cigar * 0 0 sequence_ref qual NM: MD:
        # AS XS RG ....

        fout = open(self.outfile, "w")
        fout.write("@HD\tVN:1.3\tSO:unsorted\n")

        msg = "maf2paf found q starting a new line."
        with open(self.filename, "r") as fin:
            # scan once to get sequence name and length.
            # only reference is of interest
            names = set()
            top = True
            for line in fin:
                if len(line.strip()) == 0:
                    top = True
                    continue

                m = MAFLine(line)
                if m.mode == "s" and top is True:
                    if m.name not in names:
                        names.add(m.name)
                        fout.write(
                            "@SQ\tSN:{}\tLN:{}\n".format(m.name, m.sequence_size)
                        )
                    top = False

        from bioconvert import version

        fout.write(
            "@PG\tID:{0}\tPN:{0}\tVN:{1}\tCL:{0} {2} {3}\n".format(
                "bioconvert", version, self.filename, self.outfile
            )
        )

        with open(self.filename, "r") as fin:
            for line in fin:
                # skipping empty lines
                if line.strip() == "":
                    continue

                if line[0] == "a":
                    s = []
                    if "=" in line:
                        tags = dict(i.split("=") for i in line[1].split())
                    else:
                        tags = {}
                elif line[0] == "s":
                    s.append(line)
                elif line[0] == "q":
                    # quality ?
                    raise NotImplementedError(msg)
                elif line[0] == "p":
                    raise NotImplementedError(msg)
                elif this[0] == "#":
                    print(this)

                # now that we have the two lines, save into SAM file
                if len(s) > 2:
                    raise NotImplementedError("mutliple alignment not implemented yet")

                if len(s) == 2:
                    ref = MAFLine(s[0])
                    query = MAFLine(s[1])

                    if ref.strand != "+":
                        raise Exception(
                            "for SAM, the 1st strand in each alignment must be +"
                        )

                    flag = self.get_flag(ref.alignment, query.strand)

                    # This is the slowest part of the code
                    cigar = get_cigar(ref, query)

                    qual = "*"

                    if "mismap" in tags:
                        mapq = tags["mismap"]
                    else:
                        mapq = "255"  # missing  254 is maximum

                    pos = int(ref.alignment_start) + 1  # convert to 1-based coordinate
                    data = [
                        query.name,
                        flag,
                        ref.name,
                        pos,
                        mapq,
                        cigar,
                        "*",
                        0,
                        0,
                        query.alignment.replace("-", "").upper(),
                        qual,
                    ]
                    # MD
                    # XS
                    # RG:Z:1
                    if "score" in tags:
                        data.append("AS:i:{}".format(tags["score"]))

                    if "expect" in tags:
                        data.append("EZ:Z:{}".format(tags["expect"]))

                    # try: mapq = mapqFromProb(maf.namesAndValues["mismap"])
                    # except KeyError: mapq = mapqMissing
                    editDistance = sum(
                        1 for x, y in zip(ref.alignment, query.alignment) if x != y
                    )
                    # no special treatment of ambiguous bases: might be a minor bug
                    editDistance = "NM:i:" + str(editDistance)

                    data.append(editDistance)

                    fout.write("\t".join([str(x) for x in data]) + "\n")
                    s = []

        if len(s):
            raise ValueError("Your MAf file seems truncated. ")

        fout.close()

    def get_flag(self, qName, query_strand):
        if qName.endswith("/1"):
            qName = qName[:-2]
            if query_strand == "+":
                flag = "99"  # 1 + 2 + 32 + 64
            else:
                flag = "83"  # 1 + 2 + 16 + 64
        elif qName.endswith("/2"):
            qName = qName[:-2]
            if query_strand == "+":
                flag = "163"  # 1 + 2 + 32 + 128
            else:
                flag = "147"  # 1 + 2 + 16 + 128
        else:
            if query_strand == "+":
                flag = "0"
            else:
                flag = "16"
        return flag


def mapqFromProb(probString):
    mapqMaximum = 100
    try:
        p = float(probString)
    except ValueError:
        raise Exception("bad probability: " + probString)
    if p < 0 or p > 1:
        raise Exception("bad probability: " + probString)
    if p == 0:
        return mapqMaximum
    phred = -10 * math.log(p, 10)
    if phred >= mapqMaximum:
        return str(mapqMaximum)
    return str(int(round(phred)))


def cigarCategory(alignmentColumn):
    x, y = alignmentColumn
    if x == "-":
        if y == "-":
            return "P"
        else:
            return "I"
    else:
        if y == "-":
            return "D"
        elif x != y:
            return "X"
        else:
            return "M"


def cigarParts(beg, alignmentColumns, end):
    if beg:
        yield str(beg) + "H"
    # (doesn't handle translated alignments)
    for k, v in groupby(alignmentColumns, cigarCategory):
        yield str(sum(1 for _ in v)) + k
    if end:
        yield str(end) + "H"


def get_cigar(m1, m2):
    qRevStart = m2.sequence_size - m2.alignment_start - m2.alignment_size
    cigar = "".join(
        cigarParts(m2.alignment_start, zip(m1.alignment, m2.alignment), qRevStart)
    )
    return cigar



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018-2020  Institut Pasteur, Paris and CNRS.                #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
""" Convert :term:`FASTA` (scaffold) to :term:`FASTA` (contig) and :term:`AGP` formats"""
from bioconvert.core.base import ConvArg
from bioconvert import ConvBase
from bioconvert import requires
from bioconvert.core.decorators import requires, requires_nothing

from math import log10

__all__ = ["FASTA2FASTA_AGP"]


class FASTA2FASTA_AGP(ConvBase):
    """Convert :term:`FASTA` file of scaffolds to a FASTA file of
    contigs and an :term:`AGP` file

    Method implemented in Python by bioconvert developers.

    Possible option for the future version:

    - scaffolds shorter than this length will be excluded (200)
    - scaftigs shorter than this length will be masked with "N"s (50)
    - minimum length of scaffolding stretch of Ns (default 10) to split
      scaffold into several contigs

    if input sequence is on several lines, the output contig file
    will save the sequence on a single line

    converts to upper cases

    version 2.0 (columns 9 not empty)
    columns are:


    * Column 1 contains the object, that is the identifier for the object
      assembled. This can be a chromosome, scaffold or contig. If an accession
      version number is not used, convention is to preceded a chromosome with *chr*
      (e.g. chr1) and linkage group numbers with *LG*. Contigs and scaffolds may have
      identifier that is unique within the assembly.
    * Columns 2 contains the *object_beg*: The starting coordinates of the
      component/gap on the object in column 1. These are the location in the object’s
      coordinate system, not the component’s.
    * Columns 3 is the *object_end*: The ending coordinates of the component/gap
      on the object in column 1. These are the location in the object’s coordinate
      system, not the component’s.
    * Columns 4 is the *part_number*: The line count for the components/gaps
      that make up the object described in column 1.
    * Columns 5 component_type: The sequencing status of the component. These
      typically correspond to keywords in the International Sequence Database
      (GenBank/EMBL/DDBJ) submission. Some acceptable values used here: W (WGS
        contig), N (gap with specified size), U (gap of unknown size)
    * columns 6a: component_id
    * columns 6b: gap_length
    * columns 7a: component_beg:
    * columns 7b: gap_type: if columns 5 in {N,U} specifices gap type:
        - scaffold
        - contig
        - centromere
        - short_arm
        - telomere
        - repeat
        - contamination
    * columns 8a: component_end
    * columns 8b: linkage (if column equal to N or U): tells if there is
      evidence of linkage between the adjacent lines (yes or no)
    * Columns 9a orientation: if column 5 not in {N, U}, specifies orientation
      of object in column1: + (plus), - (minus), ? (unknown), 0 (zero), na
      (irrelevant). ?, 0 and na are deprecated. use + instead (default)
    * Columns 9b linkage evidence: if column 5 is in {N,U}, linkage accepted
      values can be diverse:
        - paired-ends (paired sequences from the two ends of
          a DNA fragment, mate-pairs and molecular-barcoding.), 
        - 

https://github.com/bcgsc/abyss/blob/master/bin/abyss-fatoagp

    """

    #: Default value
    _default_method = "python"
    min_scaffold_length = 200
    min_stretch_of_Ns = 10

    def __init__(self, infile, outfile, *args, **kargs):
        """.. rubric:: constructor

        :param str infile: input FASTA file
        :param str outfile: output AGP filename

        """
        super(FASTA2FASTA_AGP, self).__init__(infile, outfile)
        self.outfile_fasta = outfile[0]
        self.outfile_agp = outfile[1]

    def _mask_scatigs(self, x, min_scatigs=10):
        # scaftigs shorter than this length will be masked with "N"s
        # AAAAAAAAAAAAA-NNNNN-GG-NNNNNN-AAAAAAAAAAAAA
        # Here the GG will be masked.
        def tt():
            if len(x) == 0:
                return "N"
            elif len(x) <= min_scatigs:
                return len(x) * "N"
            else:
                return x

        return "".join([tt(x) for x in f.split("N")])

    @requires_nothing
    def _method_python(self, *args, **kwargs):
        """Converts the input FASTA (scaffold) into FASTA (contigs) and AGP. Internal method"""
        min_scaffold_length = kwargs.get(
            "min_scaffold_length", self.min_scaffold_length
        )
        stretch_of_Ns = kwargs.get("min_stretch_of_Ns", self.min_stretch_of_Ns)
        stretch_of_Ns = "N" * stretch_of_Ns

        # First, we need to figure out number of sequences
        counter = 0
        with open(self.infile, "r") as fin:
            for line in fin.readlines():
                if line.startswith(">"):
                    counter += 1

        # a simple lambda function
        ZFILL = int(log10(counter)) + 1

        def _frmt(counter):
            return "{}".format(str(counter).zfill(ZFILL))

        # We scan the input scaffold file and create (1) the contig fasta file
        # and (2) the AGP file.
        scaffold_counter = 0
        contig_counter = 0

        # given a scaffold AAANNNCCCNNNTTT :
        # This is made of 3 contigs (AAA, CCC and TTT)
        # The 2 stretch of Ns are scaffold. The linkage is paired-ends

        # NNNN should be trimmed

        # mask scaftigs shorter than -S threshold with "N"s

        with open(self.infile, "r") as fin:
            fout_fasta = open(self.outfile_fasta, "w")
            fout_agp = open(self.outfile_agp, "w")
            fout_agp.write("##agp-version   2.0\n")

            for line in fin.readlines():
                if line.startswith(">") and scaffold_counter != 0:
                    # a new line and we already read the first sequence.
                    # So, so we need to save the contig now and the
                    # AGP information into a file
                    fout_fasta.write(">contig_" + _frmt(scaffold_counter) + "\n")
                    fout_fasta.write("{}\n".format(current))

                    # Save the previous AGP line if required
                    L = len(current)
                    data = [
                        "scaffold_" + _frmt(scaffold_counter),
                        1,
                        L,
                        1,
                        "W",
                        "contig_" + _frmt(contig_counter),
                        1,
                        L,
                        "+",
                    ]
                    data = [str(x) for x in data]
                    fout_agp.write("\t".join(data) + "\n")

                    # reset the current sequence
                    current = ""

                    # now we increment the counter
                    scaffold_counter += 1

                elif line.startswith(">") is False:
                    # accumulating the sequence with upper case
                    current += line.strip().upper()
                    # Do we have Ns
                    indexN = current.find(stretch_of_Ns)
                    if indexN != -1:
                        print("Found Ns")

                elif line.startswith(">"):
                    # This is the first line of the file
                    current = ""
                    scaffold_counter += 1
                else:
                    raise IOError("Not a valid FASTA file")

            # finally, the last line in the AGP file
            fout_fasta.write(">contig_" + _frmt(scaffold_counter) + "\n")
            fout_fasta.write("{}\n".format(current))
            L = len(current)
            data = [
                "scaffold_" + _frmt(scaffold_counter),
                1,
                L,
                1,
                "W",
                "contig_" + _frmt(contig_counter),
                1,
                L,
                "+",
            ]
            data = [str(x) for x in data]
            fout_agp.write("\t".join(data) + "\n")
            fout_agp.close()
            fout_fasta.close()

    @classmethod
    def get_additional_arguments(cls):
        yield ConvArg(
            names="--min-scaffold-length",
            default=FASTA2FASTA_AGP.min_scaffold_length,
            type=int,
            help="minimum scaffold length",
        )
        yield ConvArg(
            names="--min-stretch-of-N",
            default=FASTA2FASTA_AGP.min_stretch_of_Ns,
            type=int,
            help="minimum stretch of Ns to split a scaffold",
        )

    @staticmethod
    def get_IO_arguments():
        yield ConvArg(
            names="input_file",
            default=None,
            type=ConvArg.file,
            help="Path to the input scaffold FASTA file.",
        )
        yield ConvArg(
            names="output_file",
            nargs=2,
            default=None,
            type=ConvArg.file,
            output_argument=True,
            help="contig FASTA file followed by the AGP file.",
        )



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Tools for benchmarking"""

import statistics

import colorlog
import matplotlib.pyplot as plt
import pandas as pd
import pylab
import statsmodels.api
import statsmodels.formula.api
import statsmodels.stats.multitest
from easydev import Timer
from tqdm import tqdm

_log = colorlog.getLogger(__name__)


__all__ = ["Benchmark", "plot_multi_benchmark_max"]


class Benchmark:
    """Convenient class to benchmark several methods for a given converter

    ::

        c = BAM2COV(infile, outfile)
        b = Benchmark(c, N=5)
        b.run_methods()
        b.plot()

    """

    def __init__(self, obj, N=5, to_exclude=None, to_include=None):
        """.. rubric:: Constructor

        :param obj: can be an instance of a converter class or a class name
        :param int N: number of replicates
        :param list to_exclude: methods to exclude from the benchmark
        :param list to_include: methods to include ONLY

        Use one of *to_exclude* or *to_include*.
        If both are provided, only the *to_include* one is used.

        """
        if isinstance(obj, str):
            raise NotImplementedError

        self.converter = obj
        self.N = N
        self.results = None
        if to_exclude is None:
            self.to_exclude = []
        else:
            self.to_exclude = to_exclude
        if to_include is None:
            self.to_include = []
        else:
            self.to_include = to_include

    def run_methods(self):
        """Runs the benchmarks, and stores the timings in *self.results*."""
        results = {}
        methods = self.converter.available_methods[:]  # a copy !

        if self.to_include:
            methods = [x for x in methods if x in self.to_include]
        elif self.to_exclude:
            methods = [x for x in methods if x not in self.to_exclude]

        for method in methods:
            times = []
            for i in tqdm(range(self.N), desc="Evaluating method {}".format(method)):
                with Timer(times):
                    # Need to get all extra arguments for specify method e.g Bam2BIGWIG.uscs method
                    kwargs = {"method": method}
                    for k, v in self.converter.others.items():
                        kwargs[k] = v
                    self.converter(**kwargs)
            results[method] = times
        self.results = results

    def plot(self, rerun=False, ylabel="Time (seconds)", rot_xticks=0, boxplot_args={}):
        """Plots the benchmark results, running the benchmarks
        if needed or if *rerun* is True.

        :param rot_xlabel: rotation of the xticks function
        :param boxplot_args: dictionary with any of the pylab.boxplot arguments
        :return: dataframe with all results
        """

        if self.results is None or rerun is True:
            self.run_methods()

        # an alias.
        data = self.results.copy()

        methods = sorted(data, key=lambda x: pylab.mean(data[x]))
        pylab.boxplot([data[x] for x in methods], **boxplot_args)
        # pylab.xticks([1+this for this in range(len(methods))], methods)
        if "vert" in boxplot_args and boxplot_args["vert"] is False:
            pylab.yticks(*zip(*enumerate(methods, start=1)), rotation=rot_xticks)
            pylab.xlabel(ylabel)
        else:
            pylab.xticks(*zip(*enumerate(methods, start=1)), rotation=rot_xticks)
            pylab.ylabel(ylabel)
            pylab.xlim([0, len(methods) + 1])

        pylab.grid(True)
        pylab.tight_layout()

        return data


def plot_multi_benchmark_max(path_json, output_filename="multi_benchmark.png"):
    """Plotting function for the Snakefile_benchmark to be found in the doc

    The json file looks like::


        {
          "awk":{
            "0":0.777020216,
            "1":0.9638044834,
            "2":1.7623617649,
            "3":0.8348755836
          },
          "seqtk":{
            "0":1.0024843216,
            "1":0.6313509941,
            "2":1.4048073292,
            "3":1.0554351807
          },
          "Benchmark":{
            "0":1,
            "1":1,
            "2":2,
            "3":2
          }
        }

    Number of benchmark is infered from field 'Benchmark'.

    """
    # open and read JSON file
    df = pd.read_json(path_json)

    # how many runs per method ?
    N = len(df['Benchmark']) / len(df['Benchmark'].unique())
    N = len(df['Benchmark']) / N

    # Retrieving method names
    methods = list(df)

    # Removed the entry from the list that matches benchmark
    methods.remove('Benchmark')

    # We rotate the JSON object in relation to the benchmark number to be able to group them by methods
    df2 = df.pivot(columns="Benchmark")

    # Display of the boxplots with font at 7 and the grid is removed to make it easier to read
    df2.boxplot(fontsize=7, grid=False)

    # Initializing the x-axis title placement list
    l = []
    # Initialization of the variable which will be used to separate the different methods
    sep = 0
    # Initialization of the variable that will be used to save the lowest median
    median_best = 0
    # Initialization of the variable that will be used to save the name of the method with the lowest median
    best_method = None
    for i in methods:
        if sep != 0:
            # Creation of a separation between methods
            plt.axvline(sep + 0.5, ls="--", color="red")
        # Calculation of the median for a method
        median = statistics.median(df[i])
        if median_best == 0 or median_best >= median:
            # If the calculated median is lower than the recorded one, the old median and the old method are placed by the new one
            median_best = median
            best_method = i
        # We plot the median of each method
        plt.hlines(y=median, xmin=0.5 + sep, xmax=N+0.5 + sep, color="orange")
        l.append(sep + (N/2+0.5))
        sep += N

    # The name of each method is displayed on the x-axis
    plt.xticks(l, methods)
    # Creation of the path variable which will give the title of the output PNG image
    path = path_json.split("/")[1]
    # Backup of the benchmark of the different conversion in the form of a PNG image
    plt.savefig(output_filename, dpi=200)

    ############################## T-TEST ##############################
    # We recover the different times of the best method
    value_best_method = df[best_method]
    # Initialization of the dictionary which will save the results of the t-test of each method
    t_test = dict()
    for i in methods:
        if i != best_method:
            # We recover the different times of the method
            value_method = df[i]
            # Application of the t-test between the best method and all the other methods and saving these results in the dictionnary t_test
            comp = statsmodels.stats.weightstats.CompareMeans.from_data(
                value_best_method, value_method
            )
            (T_stats, P_value, degrees_f) = comp.ttest_ind()
            T_dict = {"t-stats": T_stats}
            P_dict = {"p-value": P_value}
            D_dict = {"Degree of freedom": degrees_f}
            t_test[i] = (T_dict, P_dict, D_dict)

    ############################## MULTITEST ##############################
    # Initialization of the list which will store all the p-values of the previous t-tests
    list_p_value = []
    # Initialization of the list which will store all the methods other than the best method
    list_method = []

    for i in t_test:
        # Retrieval of the p-values and the name of the different methods
        list_p_value.append(t_test[i][1]["p-value"])
        list_method.append(i)

    # Application of the multitest on the different conversion methods using the bonferroni method
    (
        areSignificant,
        correctedPvalues,
        _,
        _,
    ) = statsmodels.stats.multitest.multipletests(
        list_p_value, alpha=0.05, method="bonferroni"
    )

    for i in range(len(list_method)):
        print(
            f"- By comparing the {list_method[i]} method with the best one ({best_method}), we check H0: {areSignificant[i]} with corrected P-value: {correctedPvalues[i]}"
        )



# -*- coding: utf-8 -*-

###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################

import re


class Gff2:
    """Read a gff v3 file
	See the format description at https://github.com/The-Sequence-Ontology/Specifications/blob/master/gff3.md
	"""

    def __init__(self, filename):
        self.filename = filename

    def read(self):
        """ Read annotations one by one creating a generator """

        with open(self.filename) as reader:
            line = None

            for line in reader:
                # Format checking
                split = line.rstrip().split("\t")
                if len(split) < 9:
                    # Wrong line format
                    if len(split) > 0:
                        print(
                            "Impossible to read the following line regarding the gff3 specifications"
                        )
                        print(line)
                    continue

                annotation = self.process_main_fields(split[0:8])
                annotation["attributes"] = self.process_attributes(split[8])

                yield annotation

    def process_main_fields(self, fields):
        annotation = {}

        # Unique id of the sequence
        annotation["seqid"] = fields[0]
        # Optional source
        if fields[1] != ".":
            annotation["source"] = fields[1]
        # Annotation type
        annotation["type"] = fields[2]
        # Start and stop
        annotation["start"] = int(fields[3])
        annotation["stop"] = int(fields[4])
        # Optional score field
        if fields[5] != ".":
            annotation["score"] = float(fields[5])
        # Strand
        if fields[6] == "+" or fields[6] == "-" or fields[6] == "?" or fields[6] == ".":
            annotation["strand"] = fields[6]
        # Phase
        if fields[7] != ".":
            annotation["phase"] = int(fields[7]) % 3

        return annotation

    def process_attributes(self, text):
        attributes = {}

        # split into mutliple attributes
        split = text.rstrip().split(";")
        for attr in split:
            if len(attr) == 0 or attr == ".":
                continue

            # Remove parasite spaces
            attr = attr.strip()

            # Find the key and value separator
            idx = attr.find(" ")
            # split key from value
            key, value = attr[:idx].strip(), attr[idx + 1 :].strip()
            # format the value
            if value[0] == '"':
                value = value[1:-1]

            if not key in attributes:
                attributes[key] = value
            else:
                if not isinstance(attributes[key], list):
                    attributes[key] = [attributes[key]]
                attributes[key].append(value)

        return attributes


# if __name__ == "__main__":
# 	file = "/home/yoann/Projects/Hub/bioconvert/bioconvert/data/gff2_example.gff"
# 	gff = Gff2(file)
# 	for annotation in gff.read():
# 		print(annotation)



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`EMBL` file to :term:`FASTA` format"""
from bioconvert import ConvBase
from bioconvert.core.decorators import requires
from bioconvert.core.decorators import compressor


__all__ = ["EMBL2FASTA"]


class EMBL2FASTA(ConvBase):
    """Convert :term:`EMBL` file to :term:`FASTA` file

    Methods available are based on squizz [SQUIZZ]_ or biopython
    [BIOPYTHON]_.
    """

    #: Default value
    _default_method = "biopython"

    def __init__(self, infile, outfile, *args, **kargs):
        """.. rubric:: constructor

        :param str infile: input EMBL file
        :param str outfile: output FASTA filename

        """
        super(EMBL2FASTA, self).__init__(infile, outfile, *args, **kargs)

    # did not work on example
    @requires("squizz")
    @compressor
    def _method_squizz(self, *args, **kwargs):
        """Header is less informative than the one obtained with biopython
        
        """
        cmd = "squizz  -f embl -c fasta {} > {} ".format(self.infile, self.outfile)
        self.execute(cmd)

    @requires(python_library="biopython")
    @compressor
    def _method_biopython(self, *args, **kwargs):
        """For this method we use the biopython package Bio.SeqIO. 
        
        `Bio.SeqIO Documentation <https://biopython.org/docs/1.76/api/Bio.SeqIO.html>`_"""
        from Bio import SeqIO

        SeqIO.convert(self.infile, "embl", self.outfile, "fasta")



###########################################################################
# Bioconvert is a project to facilitate the interconversion               #
# of life science data from one format to another.                        #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2018  Institut Pasteur, Paris and CNRS.                     #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# bioconvert is free software: you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# bioconvert is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
"""Convert :term:`VCF` format to :term:`WBPLINK` formats"""

import os
import colorlog

from bioconvert import ConvBase
from bioconvert.core.decorators import requires
from bioconvert.core.utils import generate_outfile_name

_log = colorlog.getLogger(__name__)


class VCF2BPLINK(ConvBase):
    """Converts a genotype dataset vcf :term:`VCF` format to
    bed+bim+fam in :term:`BPLINK` format

    Conversion is based on plink executable

    """

    #: Default value
    _default_method = "plink"

    def __init__(self, infile, outfile=None, *args, **kwargs):
        """.. rubric:: constructor

        :param str infile: input :term:`PLINK` file.
        :param str outfile: (optional) output :term:`BPLINK` file
        """
        if not outfile:
            outfile = os.path.splitext(infile)[0]
        super(VCF2BPLINK, self).__init__(infile, outfile)

    @requires("plink")
    def _method_plink(self, *args, **kwargs):
        """Convert using plink executable.

        `plink documentation <http://hpc.ilri.cgiar.org/beca/training/data_mgt_2017/BackgroundMaterial/PlinkTutorial.pdf>`_"""
        cmd = "plink --vcf {infile} --make-bed --out {outfile}".format(
            infile=self.infile, outfile=self.outfile
        )
        self.execute(cmd)
